//-------------------------------------------------------------------//
//                                                                   //
//  JFC Copyright (C) Glenn Rosenthal, 1999,2000,2001.               //
//  All rights reserved.                                             //
//                                                                   //
//-------------------------------------------------------------------//

#ifndef jfc_conf_h
#define jfc_conf_h

#define SIZE_NAME           40  // Space allocated for the name of fonts

#define OPENREAD(f)     CreateFile (f,GENERIC_READ ,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL)
#define OPENWRITE(f)    CreateFile (f,GENERIC_WRITE,0,NULL,OPEN_ALWAYS  ,FILE_ATTRIBUTE_NORMAL,NULL)

                                // File open flags
#define OPEN_READ           0   // Open file for read
#define OPEN_WRITE          1   // Open file for write
#define OPEN_APPEND         2   // Open file for append

                                // Color kanji flags
#define COLORKANJI_OFF      0   // Disables the kanji font coloring.
#define COLORKANJI_MATCH    1   // Colors kanji that match the list.
#define COLORKANJI_NOMATCH  2   // Colors kanji that don't match the list.

                                // Options for dealing with an exit with an open file
#define EXIT_ASK            0   // Ask user
#define EXIT_QUIT           1   // Quiet exit
#define EXIT_SAVE           2   // Save the file

                                // Options for dealing with a double click.
#define DOUBLE_INFO         0   // Double click does a character information
#define DOUBLE_RIGHT        1   // Double click shows the card
#define DOUBLE_DISABLE      2   // Disables the double click

#ifdef WINCE
  #define CONFIG_MAGIC  0x1d72C532  // Magic ID for JFC config files.
#else
  #define CONFIG_MAGIC  0x2d72C532  // Magic ID for JFC config files.
#endif WINCE

struct cfg_font {
  TCHAR name[SIZE_NAME];        // Name of font
  short size;                   // Size of font
  byte  automatic;              // Font is automatically sized.
  byte  cache;                  // Extra byte for who knows what.
};

struct cfg {
  long  magic;                  // Identifies this as a JFC config file.
  byte  install;                // If set causes check for installed version and file extensions.
  byte  status;                 // Generate a status bar
  byte  short_name;             // Use short names in the title bar.
  byte  maximize;               // Maximaize the file.
  byte  usedims;                // Use last saved dimensions.
  byte  save_exit;              // Save configuration on exit.     
  byte  reload_files;           // Reload files loaded when we exited.
  byte  exit_open_file;         // Exit quietly even if user is in a card set.
  byte  flash_kanji;            // Show kanji and let user guess kana & meaning
  byte  flash_kana;             // Show kana and let user guess kanji & meaning
  byte  flash_ascii;            // Show ascii and let user guess kana & kanji
  byte  kanji_kanji;            // Show kanji and let user guess readings & meaning (kanji cards)
  byte  kanji_kana;             // Show readings and let user guess kanji & meaning (kanji cards)
  byte  kanji_ascii;            // Show ascii and let user guess readings & kanji (kanji cards)
  byte  kanji_always;           // Always show kanji with kana.
  byte  kanji_unknown;          // Only show kanji witht he kana when it is unknown
  byte  know_all_kanji;         // Know all kanji
  byte  full_kanjicard;         // Show advanced kanji cards as full headers
  byte  duplicate_cards;        // Duplicate cards that are missed.
  byte  double_click;           // Determines action when a double click is detected.
  byte  multi_mode;             // Display mode used for multi mode.
  byte  knownkanji_colorkanji;  // Assume that color kanji indicates the known kanji.
  byte  colorkanji_mode;        // Determines the way color-fonts are suported.
  byte  swap_mouse;             // Swap elements on mouse down/up
  byte  review_mode;            // Use review mode and the weights.
  byte  keep_stats;             // Keep statistics.
  byte  stat_config;            // Use configuation from status file.
  byte  no_move_end;            // Do not place wrong card at the end.
  byte  time_cards;             // If user does not answer in time the card is wrong.
  byte  use_max;                // Limit number of cards generated.
  byte  all_fonts;              // Show all fonts.
  byte  cache_displayfont;      // Should we cache or not cache the display font.
  byte  cache_systemfont;       // Should we cache or not cache the system font.
  byte  kana_for_advanced;      // Uses kana font for advanced kanji compounds.
  byte  info_titles;            // Puts titles in the kanji-info list box.
  byte  info_compress;          // Compress Character information.
  byte  cache_info;             // Fill for later exapnsion
  byte  wheel_control;          // Uses the wheel control to control the program (PPC only)
  byte  one_click;              // Use one click simplified controls
  byte  auto_expand;            // Automatically expand shown items when moving to the next item.
  byte  auto_collapse;          // Automatically collapse shown items when moving to next item
  COLORREF colorkanji_color;    // Color to be used with color-kanji.
  COLORREF info_color;          // Color used for titles in kanji-info box.
  int   x,y,xs,ys;              // Dimensions of last saved configuration.
  struct cfg_font ascii_font;   // Ascii font.
  struct cfg_font kanji_font;   // Kanji font.
  struct cfg_font kana_font;    // Kana font
  struct cfg_font imi_font;     // meaning font.
  struct cfg_font sys_font;     // System font.
  struct cfg_font big_font;     // Big font.
  struct cfg_font extra_font;   // Extra font for later.
  short font_cache;             // Size of font cache in characters.
  short duplicate_count;        // Frequency for duplicated cards.
  int   maximum;                // Maximum number of cards generated.
  int   time_show;              // Time to show the card in msec.
  int   time_next;              // Time to move to the next card in msec.
  int   max_subfiles;           // Maximum number of sub-files.
  byte  flashkanji[10];         // Flash kanji weights.
  byte  flashkana [10];         // Flash kana weights.
  byte  flashascii[10];         // Flash ascii weights.
  byte  kanjikanji[10];         // Kanji card kanji weight
  byte  kanjikana [10];         // Kanji card kana weight
  byte  kanjiascii[10];         // Kanji card ascii weight
  byte  fill[128];              // Filler for later expansion
};

class JFC_config {
public:
  struct cfg cfg;           // Most of the actual data
  TCHAR *load;              // Pointer to the data on previously loaded files.
private:
  byte   ok;                // Disk based configuration is valid.
  TCHAR  buffer[256];       // Name of configuration file.
  TCHAR *ptr;               // Pointer to end of buffer base.
  void  set          (struct cfg *new_config);                      // Set a configuration.
public:
  JFC_config         (void);                                        // Constructor.
  void  def          (void);                                        // Set to default.
  void  done         (void) { if (!ok || cfg.save_exit) write(); }  // Called during exit to write configuration.
  HANDLE open        (TCHAR *name,int mode);                        // Open a file based on configurations settings.
  TCHAR *name        (TCHAR *file);                                 // Generate file name.
  TCHAR inline *name (void) { return (buffer); }                    // Get last name generated by program.
  void  options      (void);                                        // Do options.
  int   read         (void);                                        // Reader routine.
  void  write        (void);                                        // Write configuration file
  void  write_files  (void);                                        // Write files information to the configuation file.
};

extern class JFC_config jfc_config;                 // Class instance.

extern int initialize_config (WNDCLASS *wclass);    // Initialize window classes used in the Options dialog.

#endif jfc_conf_h
