//-------------------------------------------------------------------//
//                                                                   //
//  JFC Copyright (C) Glenn Rosenthal, 1999,2000,2001.               //
//  All rights reserved.                                             //
//                                                                   //
//-------------------------------------------------------------------//

//-------------------------------------------------------------------
//
//  This module implements japanese versions of the windows controls.
//  In particular, this module implements a japanese text box control
//  This control is used to display most text in this program.
//

#include "jfc_main.h"
#include "jfc_conf.h"
#include "jfc_edit.h"
#include "jfc_file.h"
#include "jfc_font.h"
#include "jfc_info.h"
#include "jfc_misc.h"
#include "jfc_wind.h"

//-------------------------------------------------------------------
//
//  Exported structures.
//

SCROLLINFO scroll_info = { sizeof(SCROLLINFO),SIF_ALL | SIF_DISABLENOSCROLL,0,0,0,0,0 };    

//===================================================================
//
//  JFC Edit routines.  These routines control the display box used to
//  display Japanese text.  The term edit is not really correct and 
//  should have be text box.
//
//

//
//  These define the width widths being added to the size of the windows by Windows.
//  The PPC systems are using a simple flat view that is more like the PocketPC view.
//  This is really an old look for Windows.
//
#ifdef WINCE_PPC
  #define TEXT_XEDGE GetSystemMetrics(SM_CXBORDER)
  #define TEXT_YEDGE GetSystemMetrics(SM_CYBORDER)
#else  WINCE_PPC
  #define TEXT_XEDGE GetSystemMetrics(SM_CXEDGE)
  #define TEXT_YEDGE GetSystemMetrics(SM_CYEDGE)
#endif WINCE_PPC

//-------------------------------------------------------------------
//
//  Static routines.
//

static LRESULT CALLBACK JWP_text_proc (HWND hwnd,UINT iMsg,WPARAM wParam,LPARAM lParam);

//
//  This is the window proc for the japanes text box.  This is the big
//  deal.  
//
//  Special messages:
//
//      JT_SETTEXT    -- Sets the text buffer of the edit box.  
//                       wParam -- Length,  lParam -- Pointer to kanji string.
//
static LRESULT CALLBACK JFC_text_proc (HWND hwnd,UINT iMsg,WPARAM wParam,LPARAM lParam) {
//
//  Creation, we need to create the JWP_file object, and adjust the size
//  of the window to refelect the actual size of the kanji font.
//
  if (iMsg == WM_CREATE) SetWindowLong (hwnd,0,(long) ((CREATESTRUCT *) lParam)->lpCreateParams);       // Save refference to our structure.
//
//  Service all other messages through the class.
//
  JFC_text *text;
  text = (JFC_text *) GetWindowLong(hwnd,0);    // Get our JWP_fiel class object.
  return (text->win_proc(hwnd,iMsg,wParam,lParam));
}

//-------------------------------------------------------------------
//
//  Begin class JFC_Text
//
//  This class manages the text windows that as associated with lines
//  of Japanese text.
//

static int   x;         // Static location being used for formatting text lines.
static short timer;     // Static location for timer flag.  This has to be gobal because we
                        //   want the wheel options to work.  These required this to be 
                        //   global because when we do a double ckick on the up or down keys,
                        //   we can't have this flag set in each display.  The value set in
                        //   the flag is the scan code for the key.  This prvents up/down from
                        //   recording as as a second button.
//
//  Constructor.
//
JFC_text::JFC_text () {
  window = NULL;
  buffer = NULL;
  line2  = NULL;
  line3  = NULL;
  line4  = NULL;
  first  = NULL;
  font   = &kana_font;
  clear ();
  return;
}

//
//  Destructor.
//
JFC_text::~JFC_text () {
  clear ();
  return;
}

//
//  Adjust the position of the window.  We do not reformat this routine.  Generally this is
//  called as a responce to a chainge of viewing informat and we want to redraw all the 
//  windows.
//
//      x,y -- Location of new window.
//
void JFC_text::adjust (int x,int y) {
  MoveWindow (window,x,y,width,height,true);
  draw       ();
  return;
}

//
//  Deallocate resources and place item in defined state.
//
void JFC_text::clear () {
  mode       = MODE_SINGLE;
  view_state = false;
  if (buffer) free (buffer);
  if (line2 ) free (line2);
  if (line3 ) free (line3);
  if (line4 ) free (line4);
  format ();                    // Trick to deallocate the line buffers, is to fromat to nothing.
  return;
}

//
//  Simple fucntion to do all the work associated with getting character 
//  infrormation.  This includes extracting the character from the mouse
//  click location.
//
//      lParam -- Contains mouse location inforamtion.
//
void JFC_text::do_info (LPARAM lParam) {
  int x,i,j;
  struct JFC_line *line;
  KANJI_info info;
  if (!view_state) return;                          // No text is displayed.
  i = (HIWORD(lParam)-1)/font->lheight;             // Find line.
  for (j = 0, line = first; (j < i) && line->next; j++) line = line->next;
  x = LOWORD(lParam);                               // Find character
  for (i = 0, j = font->x_offset; i < line->length-1; i++) {
    j = font->hadvance(j,line->buffer[i]);
    if (j > x) break;
  }
  info.dialog (window,line->buffer[i]);             // Do information.
  return;
}

//
//  Since the double click timer was being used so much I build it into a routine.
//  This routine processes a double click event.  The event is evaluated as follows.
//
//  If the timer is acitve the menu event indicated is executed and a value of true is
//  returned.
//
//  Otherwise the timer is started and a value of false is returned.
//
//      key    -- Scan coded for key pressed.
//      menu   -- Menu event executed if this is a double click event.
//
//      RETURN -- A non-zero value indicates the event has been handled.
//
int JFC_text::do_timer (int key,int menu) {
  if (timer == key) {
    timer = false;
    jfc_file->do_menu (menu);
    return (true);
  }
  timer = key;
  SetTimer (window,TIME_KEY,GetDoubleClickTime(),NULL);
  return (false);
}

//
//  This rotuine formats the entry based on the current settings for the entry.
//
void JFC_text::format () {
  int       i,h;
  JFC_line *line;
//
//  Dispose of the current formatting.
//
  while (first) { 
    line  = first;
    first = line->next;
    delete line;
  }
//
//  Setup for the diplay modes.
//
  switch (get_mode()) {
    case MODE_SINGLE:
         if (view_state) format_string (buffer);
         break;
    case MODE_CYCLE:
         format_string ((&buffer)[multi]);
         break;
    case MODE_SWAP:
         if (!multi) format_string (buffer);
           else {
             format_string (line2);
             format_string (line3);
             format_string (line4);
           }
         break;
  }
//
//  Determine if the number of lines is changing.
//
  for (i = 0, line = first; line; line = line->next) i++;
  if (i == 0) i++;
  h      = i*font->lheight+2*TEXT_YEDGE+1;
  i      = (h == height);
  height = h;
  if (!i) jfc_window.space (id);
  return;
}

//
//  This routine adds a string to the current window display line list.  Each time this is 
//  called another line is added to the list.  This line may be broken over more than one 
//  line depending ont he width of the window.
//
//      string -- String to be added.
//
void JFC_text::format_string (KANJI *string) {
  int       i;
  if (!string) return;                              // No string do nothing.
  if (new_line(string)) return;                     // Add a line node for the string.
  while (*string) {                                 // Break lines
    x = font->hadvance(x,*string);
    if (!ISSPACE(*string) && (x >= width)) {
      if (ISJIS(*string)) {
        if (new_line(string)) return;
      }
      else {
        i = last->length;
        for (i = last->length; (i >= 0) && !ISJIS(last->buffer[i]) && !isspace(last->buffer[i]); i--);
        if (i > 0) {
          last->length = i+1;
          string = last->buffer+last->length;
        }
        if (new_line(string)) return;
      }
    }
    last->length++;
    string++;
  }
  return;
}

//
//  Get the current window mode.  This expands settings that are not resolved until display 
//  time.
//
int JFC_text::get_mode () {
  if (mode == MODE_MULTI) return (jfc_config.cfg.multi_mode);
  return (mode); 
}

//
//  Hide the window
//
void JFC_text::hide () {
  ShowWindow (window,SW_HIDE);
  return;
}

//
//  Open the window.
//
//      winnum -- ID for this window.
//
//  The window generated is not visible, and is located in the upper left hand 
//  corner of the main screen.  Generally, this would be followed by calls to 
//  move() and show() to place the window and make it visible.
//
int JFC_text::open (int winnum) {
  id = winnum;
#ifdef WINCE_PPC
  if (!(window = CreateWindow(TEXT("JFC-Text"),NULL,WS_CHILD | WS_BORDER,0,0,10,10,main_window,NULL,instance,(LPVOID) this))) return (true);
#else  WINCE_PPC
  if (!(window = CreateWindowEx(WS_EX_CLIENTEDGE,TEXT("JFC-Text"),NULL,WS_CHILD,0,0,10,10,main_window,NULL,instance,(LPVOID) this))) return (true);
#endif WINCE_PPC
  return (false);
}

//
//  Move the window
//
//      x     -- Left edge of window.
//      y     -- Top edge of window.
//      width -- Width of window.
//
//  The height of the window is not controllable.  This is determined by the 
//  contents of the winodw.
//
void JFC_text::move (int x,int y,int _width) {
  width = _width;
  format ();
  MoveWindow (window,x,y,width,height,true);
  return;
}

//
//  Add a new line to the line list.
//
//      buf    -- Pointer to the first data line in the text.
//
//      RETURN -- A non-zero value indicates we could not allocate the memory.
//
int JFC_text::new_line (KANJI *buf) {
  JFC_line *line;
  if (!(line = new JFC_line)) return (true);
  if (first) last->next = line; else first = line;
  last         = line;
  line->next   = NULL;
  line->buffer = buf;
  line->length = 0;
  x            = font->x_offset;
  return (false);
}

//
//  This sets the value contained in the line buffer for this window.
//
//      string -- Null terminated kanji string.
//      _view  -- Sets the view parameter for this window.
//      _font  -- Sets the current font for this display.
//
void JFC_text::set (KANJI *string,int _view,class JFC_font *_font) {
  font   = _font;
  mode   = MODE_SINGLE;
  buffer = kdupstr(string);
  view   (_view);
  return;
}

//
//  This routine sets a multi-display control.  The mode for the display is always set to 
//  MODE_MULTI.  This will resolve the current mode setting in the user configuration 
//  during actual use of the device.
//
//      l1-l4 -- Display lines.  Line 1 is always intiailzied at the current time.
//      _font -- Display font.
//
void JFC_text::set (KANJI *l1,KANJI *l2,KANJI *l3,KANJI *l4,class JFC_font *_font) {
  buffer = kdupstr(l1);
  line2  = kdupstr(l2);
  line3  = kdupstr(l3);
  line4  = kdupstr(l4);
  mode   = MODE_MULTI;
  multi  = 0;
  font   = _font;
  view   (true);
  return;
}

//
//  Show the window
//
void JFC_text::show () {
  ShowWindow (window,SW_SHOW);
  return;
}

//
//  Determin if the text can be seen or not.
//
void JFC_text::view (int set) {
  view_state = set;
  format ();
  draw   ();
  return;
}

//
//  Main window procedure for the text boxes.
//
int JFC_text::win_proc (HWND hwnd,int iMsg,WPARAM wParam,LPARAM lParam) {
  static TCHAR temp[] = { 0,0 };
  switch (iMsg) {
//```````````````````````````````````````````````````````````````````
//  Create
//
    case WM_CREATE:
         CREATESTRUCT *create;
         create = (CREATESTRUCT *) lParam;
         MoveWindow (hwnd,create->x,create->y,create->cx,font->height+2*font->vspace+2*TEXT_YEDGE,true);
         return (0);

//```````````````````````````````````````````````````````````````````
//
//  Render
//
    case WM_PAINT: {                              // Render.
           int         i,x,y;
           KANJI       ch;
           RECT        rect;
           HDC         hdc;
           HFONT       hfont;
           PAINTSTRUCT ps;
           JFC_line   *line;
           hdc   = BeginPaint (hwnd,&ps);
           hfont = (HFONT) SelectObject (hdc,font->ascii);
           GetClientRect (hwnd,&rect);
           SetBkMode (hdc,TRANSPARENT);
           if (active) {
             HPEN pen;
             pen = (HPEN) SelectObject(hdc,CreatePen(PS_DOT,1,RGB(0,0,0)));
             Rectangle    (hdc,rect.left,rect.top,rect.right,rect.bottom);
             DeleteObject (SelectObject(hdc,pen));
           }
           else {
             BackFillRect (hdc,&rect);
           }
           if (view_state) {                    // Text is visible.
             y = 1-font->loffset;               // Only use of loffset.
             for (line = first; line; line = line->next) {
               x =  font->x_offset;             // Setup intial position.
               y += font->lheight;
               for (i = 0; i < line->length; i++) {
                 ch = line->buffer[i];
                 if (ISJIS(ch)) font->kanji->draw(hdc,ch,x,y);
                   else if (ch != '\t') {
                     temp[0] = (TCHAR) ch;
                     TextOut (hdc,x,y-font->height,temp,1);
                   }
                 x = font->hadvance(x,ch);
               }
             }
           }
           SelectObject (hdc,font);
           EndPaint     (hwnd,&ps);
           return (0);
         }
//```````````````````````````````````````````````````````````````````
//
//  Request messages from the system
//
    case WM_GETDLGCODE:                         // We need to get input from windows.
         return (DLGC_WANTARROWS | DLGC_WANTALLKEYS | DLGC_WANTCHARS);
//```````````````````````````````````````````````````````````````````
//
//  Mouse events.
//
    case WM_LBUTTONDOWN:
         if (GetKeyState(VK_MENU) < 0) {        // Alt click is treated at a right click.
           do_info (lParam);
           return  (0);
         }
left_button_down:;
         SetFocus (hwnd);
         switch (get_mode()) {
           case MODE_SINGLE:
                if (!view_state && buffer) view (true);
                break;
           case MODE_CYCLE:
                while (true) {
                  multi++;
                  if (multi == 4) { multi = 0; break; }
                  if ((&buffer)[multi]) break;
                }
                font = (multi || ((id != 1) && jfc_config.cfg.kana_for_advanced)) ? &kana_font : &kanji_font;
                view (true);
                break;
           case MODE_SWAP:
                if (multi) multi = 0; else multi = 1;
                font = (multi || ((id != 1) && jfc_config.cfg.kana_for_advanced)) ? &kana_font : &kanji_font;
                view (true);
                break;
         }
         return (0);
    case WM_LBUTTONUP:
left_button_up:;
         if ((MODE_SWAP == get_mode()) && jfc_config.cfg.swap_mouse) {
           multi = 0;
           font = ((id != 1) && jfc_config.cfg.kana_for_advanced) ? &kana_font : &kanji_font;
           view (true);
         }
         return (0); 
    case WM_LBUTTONDBLCLK:
         switch (jfc_config.cfg.double_click) {
           case DOUBLE_INFO   : do_info (lParam); break;
           case DOUBLE_RIGHT  : jfc_file->do_menu (IDM_CARD_RIGHT); break;
           case DOUBLE_DISABLE: break;
         }
         return (0);
    case WM_RBUTTONDOWN: 
         do_info (lParam);
         return  (0);
//```````````````````````````````````````````````````````````````````
//
//  Activation events.
//
    case WM_SETFOCUS:                           // Set the focus.
         jfc_window.last_id = id;
         if (!active) {
           active = true;
           draw ();
         }
         return (0);
    case WM_KILLFOCUS:                          // Kill the focus.
         if (active) {
           if (jfc_config.cfg.auto_collapse && (MODE_SWAP == get_mode()) && (multi == 1)) {
             multi = 0;                         // Auto collapse option
             view (true);
           }
           active = false;
           draw ();
         }
         return (0);
//```````````````````````````````````````````````````````````````````
//
//  Keyboard events.
//
    case WM_KEYUP:
         switch (wParam) {
           case VK_SPACE:
           case VK_F23:
                goto left_button_up;
         }
         return (0);
    case WM_CHAR:
         if (lParam & (1L<<30)) return (0);     // Disable key repeats
         switch (wParam) {
           case '<':
           case ',':
           case 'Z':
           case 'z':
           case 'S':
           case 's':
                jfc_file->do_menu (IDM_CARD_SHOW);
                break;
           case 'X':
           case 'x':
           case '>':
           case '.':
           case ';':
           case 'R':
           case 'r':
                jfc_file->do_menu (IDM_CARD_RIGHT);
                break;
           case 'C':
           case 'c':
           case '?':
           case '/':
           case ':':
           case 'W':
           case 'w':
                jfc_file->do_menu (IDM_CARD_WRONG);
                break;
         }
         return (0);
    case WM_KEYDOWN:                            // Vitural keys.
         if (lParam & (1L<<30)) return (0);     // Disable key repeats
         int shift;
         shift = (GetKeyState(VK_SHIFT) < 0);
         switch (wParam) {
#ifndef WINCE_PPC
           case VK_RETURN:
                jfc_file->do_menu (IDM_CARD_RIGHT);
                return (0);
           case VK_ESCAPE:
                jfc_file->do_menu (IDM_CARD_WRONG);
                return (0);
#endif WINCE_PPC
           case VK_RIGHT:
           case VK_F23:
                if (jfc_config.cfg.one_click) { jfc_file->do_menu (IDM_CARD_RIGHT); return (0); }
                if (do_timer(wParam,IDM_CARD_RIGHT)) return (0);
           case VK_SPACE:
                goto left_button_down;
                return (0);
           case VK_LEFT:
           case VK_F24:
                if (do_timer(wParam,IDM_CARD_WRONG)) return (0);
                jfc_file->do_menu (IDM_CARD_SHOW);
                return (0);
           case VK_UP:
                if (jfc_config.cfg.one_click) { jfc_file->do_menu (IDM_CARD_SHOW); return (0); }
                if (jfc_config.cfg.wheel_control && do_timer(wParam,IDM_CARD_SHOW)) return (0);
                jfc_window.set (id-1,true);
                return (0);
           case VK_DOWN:
                if (jfc_config.cfg.one_click) { jfc_file->do_menu (IDM_CARD_WRONG); return (0); }
                if (jfc_config.cfg.wheel_control && do_timer(wParam,IDM_CARD_WRONG)) return (0);
                jfc_window.set (id+1,true);
                return (0);
           case VK_TAB:
                if (shift) jfc_window.set (id-1,true); else jfc_window.set (id+1,true);
                return (0);
         }
//```````````````````````````````````````````````````````````````````
//
//  Timer events.
//
    case WM_TIMER:
         if (wParam == TIME_KEY) {
           timer = false;
           KillTimer (hwnd,TIME_KEY);
         }
         return (0);
  }
  return (DefWindowProc(hwnd,iMsg,wParam,lParam));
}

//
//  End class JFC_text
//
//-------------------------------------------------------------------

//===================================================================
//
//  Routines for support of the JFC_List class.  This class manages 
//  the list used in the character information dialog.
//
//

//-------------------------------------------------------------------
//
//  Static data strucdtures.
//

//
//  This structure describes the internal memory format used by the 
//  list manager.
//
#define LIST_BLOCK  1000                // Number of lines in each block

typedef struct LIST_list {
  class  LIST_line  lines[LIST_BLOCK];  // Actual data block
  struct LIST_list *next;               // Pointer to next block.
} LIST_list;

//-------------------------------------------------------------------
//
//  Class handler fucntion
//

//
//  These define the width widths being added to the size of the windows by Windows.
//  The PPC systems are using a simple flat view that is more like the PocketPC view.
//  This is really an old look for Windows.
//
#ifdef WINCE_PPC
  #define LIST_XEDGE GetSystemMetrics(SM_CXBORDER)
  #define LIST_YEDGE GetSystemMetrics(SM_CYBORDER)
#else  WINCE_PPC
  #define LIST_XEDGE GetSystemMetrics(SM_CXEDGE)
  #define LIST_YEDGE GetSystemMetrics(SM_CYEDGE)
#endif WINCE_PPC

//
//  Window procedure for Japanese list boxes.
//
static LRESULT CALLBACK JFC_list_proc (HWND hwnd,UINT iMsg,WPARAM wParam,LPARAM lParam) {
  CREATESTRUCT *create;
  JFC_list     *list;
  int           i;        
  list = (JFC_list *) GetWindowLong(hwnd,0);        // Get our JWP_List class object.
  switch (iMsg) {
//
//  Creation, we need to create the JWP_list object, and adjust the size
//  of the window to refelect the displayable area in the window.
//
    case WM_CREATE:                                 
         create = (CREATESTRUCT *) lParam;
         i = (create->cy-2*LIST_YEDGE)/sys_font.lheight;   // Calculate viewable lines.
         MoveWindow (hwnd,create->x,create->y,create->cx,i*sys_font.lheight+2*LIST_YEDGE,true);
         list = new JFC_list (hwnd,i);
         SetWindowLong (hwnd,0,(long) list);        // Save JWP_list object for list box.
         break;
    case WM_DESTROY:                                // Destroy the JWP_list object.
         delete list;
         return (0);
    case WM_GETDLGCODE:                             // We need to get input from windows.
         return (DLGC_WANTARROWS | DLGC_WANTALLKEYS | DLGC_WANTCHARS);
  }
  return (list->win_proc(hwnd,iMsg,wParam,lParam)); // Call the class windows procedure.
}

//-------------------------------------------------------------------
//
//  begin LIST_line
//
//  This class provides some basic operations on a line containned in
//  a Japanese List box control.
//

//
//  Allocate memory and copy a string into the buffer.
//
//      len  -- Length of string to copy.
//      line -- Text to copy.
//
void LIST_line::alloc (int len,KANJI *line) {
  int i;
  clear ();                     // Clear any old value.
  if (!len) return;
  if ((text = (KANJI *) malloc(i = sizeof(KANJI)*len))) {
    memcpy (text,line,i);
    length = len;
  }
  return;
}

//
//  Deallocate all resources used by the line and crear all flags.
//
void LIST_line::clear () {
  if (text) free (text);
  length = 0;
  text   = NULL;
  return;
}

//
//  End Class LIST_line.
//
//-------------------------------------------------------------------

//-------------------------------------------------------------------
//
//  begin JFC_list class
//
//  This class provides basic control over Japanese list boxes and 
//  managed lists.
//

//
//  Constuctor for Japanese list object.
//
//      hwnd -- Window containning list.
//      l    -- Number of lines to be visible in the list
//
JFC_list::JFC_list (HWND hwnd,int l) {
  RECT rect;
  memset (this,0,sizeof(JFC_list));     // Zero out structure which sets most variables.
  GetClientRect (hwnd,&rect);           // Get client rectangle so we can set size.
  lines   = l;                          // Visible number of lines.
  width   = rect.right-rect.left;       // Display area width.
  window  = hwnd;                       // Save window pointer.
  height  = sys_font.lheight;           // Height of a display line.
  return;                               //   This is used to support dialogs with Japanese list controls.
}

//
//  Deconstructor.
//
JFC_list::~JFC_list () {
  int i;
  LIST_list *list;
  while (lists) {                       // Dispose of all lists.
    list  = lists;
    lists = list->next;
    for (i = 0; i < LIST_BLOCK; i++) list->lines[i].clear ();
    free (list);
  }
  return;
}

//
//  Put a line into the list.  If necessary the list will be extended 
//  so the line can be added.
//
//      loc  -- Line location to place this.
//      len  -- Length of the line in characters.
//      text -- Pointer to kanji text for the line.
//
void JFC_list::add_line (int len,KANJI *text) {
  LIST_line *line;
  int loc = count;
//
//  We need to allocate a new block!
//
  if (loc >= alloc) {
    LIST_list *list;
    list = (LIST_list *) calloc(1,sizeof(LIST_list));
    if (!list) return;                          // Allocation failed.
    if (!last) lists = last = list;
      else {
        last->next = list;
        last       = list;
      }
    alloc += LIST_BLOCK;                        // Correct number of allocated lines.
  }
  count++;                                      // Correct line count.
  line = get_line(loc);                         // Get the line
  line->alloc (len,text);                       // Assign string to it.
  que_line    (loc);                            // Redraw line
  scroll      ();                               // Reset scroll bar.
  return;
}

//
//  This is the main rederming routine.
//
//      hdc  -- Display context to hender to.
//      line -- Line to be rendred.
//
#define NO_CHANGE   0xFFFFFFFF

void JFC_list::draw_line (HDC hdc,int line) {
  RECT       rect;
  KANJI      ch;
  LIST_line *text;
  COLORREF   oldcolor = NO_CHANGE;
  int        i,x,y;
  static TCHAR temp[] = { 0,0 };
//
//  Generate the background rectangle.
//
  rect.top    = (line-top)*height;
  rect.bottom = rect.top+height;
  rect.left   = 0;
  rect.right  = width;
  if (focus && (line == current)) {
//    HPEN pen;
//    pen = SelectObject(hdc,CreatePen(PS_DOT,1,RGB(0,0,0)));
    Rectangle (hdc,rect.left,rect.top,rect.right,rect.bottom);
//    DeleteObject (SelectObject(hdc,pen));
  }
  else {
    BackFillRect (hdc,&rect);
  }
//
//  Render text
//
  text = get_line(line);                    // Get line
  if (!text) return;                        // No text so exit.
  x = sys_font.x_offset;                    // Setup intial position.
  y = rect.bottom-sys_font.loffset;         // Only use of loffset.
  i = 0;
//
//  Chech for highlite line.
//
  if (text->text[0] == EUC_HIGHLIGHT) {
    oldcolor = SetTextColor(hdc,jfc_config.cfg.info_color);
    i++;
  }                                                                
//
//  Render loop
//
  for (; i < text->length; i++) {
    ch = text->text[i];
    if (ISJIS(ch)) sys_font.kanji->draw(hdc,ch,x,y);
    else if (ch != '\t') {
      temp[0] = (TCHAR) ch;
      TextOut (hdc,x,y-sys_font.height,temp,1);
    }
    x = sys_font.hadvance(x,ch);
  }
//
//  Render selection indicator.
//
  if (text->selected) InvertRect(hdc,&rect);                    // Select rectangle
//
//  Cleanup
//
  if (oldcolor != NO_CHANGE) SetTextColor(hdc,oldcolor);        // Restore text color
  return;
}

//
//  Get the LIST_line structure associated with a line.  
//
//      line   -- Line number.
//
//      RETURN -- Pointer to line structure, or NULL if the line does 
//                not exist.
//
LIST_line *JFC_list::get_line (int line) {
  LIST_list *list;
  if (line >= count) return (NULL);
  for (list = lists; line >= LIST_BLOCK; list = list->next) line -= LIST_BLOCK;
  return (&list->lines[line]);
}

//
//  This is a major routine that is called any time the current position
//  in the list is changed.  
//
//      pos   -- New current position.
//
void JFC_list::move (int pos) {
  int i,j,old;
  LIST_list *list;
  
  old     = current;                        // Save old location for later
  current = pos;;                           // Calculate new location.
  if (current >= count) current = count-1;  // Clip locations to actual data.
  if (current < 0) current = 0;             //   Order is importaint when no data is there
//
//  No shift, so clear all selections and make a single selection.
//
  for (j = 0, list = lists; list; list = list->next) {
    for (i = 0; i < LIST_BLOCK; i++) {
      if (list->lines[i].selected) {
        list->lines[i].selected = false;
        que_line (j+i);                   // If selection changed que for draw.
      }
    }
    j += LIST_BLOCK;
  }
  select_count = 0;
  select (current,true);
//
//  Make sure the current line is visible.
//
  if ((current < top) || (current >= top+lines)) {      // Scroll required
    if (current < top) top = current; else top = current+1-lines;
    redraw ();
  }
  else {
    que_line (old);                         // Simple line redraw.
    que_line (current);
  }
  scroll ();                                // Adjust scroll bar
  return;
}

//
//  Marks the rectangle associated with a given line as invlaid.
//
//      line -- Line to be marked invalid.
//
void JFC_list::que_line (int line) {
  RECT rect;
  rect.top    = (line-top)*height;
  rect.bottom = rect.top+height;
  rect.left   = 0;
  rect.right  = width;
  InvalidateRect (window,&rect,false);
  return;
}

//  
//  Resets the contents of the list.  All items are reumoved and the 
//  list is redrawn.
//
void JFC_list::reset () {
  LIST_list *list;
  int        i;
  for (list = lists; list; list = list->next) {
    for (i = 0; i < LIST_BLOCK; i++) {
      list->lines[i].selected = false;
      list->lines[i].clear();
    }
  }
  count = top = current = 0;
  select_count    = 0;                  // No selected lines (no lines)
  redraw ();
  return;
}

//
//  Sets the scroll bar position for the list.
//
void JFC_list::scroll () {
  scroll_info.nMax   = count-1;
  scroll_info.nPage  = lines;
  scroll_info.nPos   = top;
  SetScrollInfo (window,SB_VERT,&scroll_info,true);
  return;
}

//
//  Change (ie set) the selection state of a given line.
//
//      line  -- Line to change the selection state of.
//      onoff -- New state.
//
void JFC_list::select (int line,int onoff) {
  LIST_line *data;
  if (!(data = get_line(line))) return;
  if ((data->selected && !onoff) || (!data->selected && onoff)) {
    if (onoff) select_count++; else select_count--;
    data->selected = onoff;
    que_line (line);
  }
  return;
}

//
//  Window procedure for Japanese list-box control
//
#define BOTTOM  (count-lines)     // Value of top when at bottom of display.

int JFC_list::win_proc (HWND hwnd,int msg,WPARAM wParam,LPARAM lParam) {
  HDC         hdc;
  HFONT       font;
  PAINTSTRUCT ps;
  int         i,j;
  switch (msg) {
//```````````````````````````````````````````````````````````````````
//
//  General window messages (create, destroy, paint).
//

//
//  Crate only generates the scroll bar
//
    case WM_CREATE:
         scroll ();
         return (0);
//
//  Set and Kill focus simply change state and redraw.
//
    case WM_SETFOCUS:
         focus = true;
         que_line (current);
         break;
    case WM_KILLFOCUS:
         focus = false;
         que_line (current);
         break;
//
//  Paint renders all lines.
//
    case WM_PAINT:
         hdc = BeginPaint (hwnd,&ps);
         font = (HFONT) SelectObject (hdc,sys_font.ascii);
         SetBkMode (hdc,TRANSPARENT);
         for (i = 0; i < lines; i++) draw_line (hdc,i+top);
         SelectObject (hdc,font);
         EndPaint (hwnd,&ps);
         return (0);
//```````````````````````````````````````````````````````````````````
//
//  Keyboard events.
//
    case WM_KEYDOWN:
         int shift,ctrl;
         shift = (GetKeyState(VK_SHIFT)   < 0);
         ctrl  = (GetKeyState(VK_CONTROL) < 0);
         switch (wParam) {
           case VK_TAB:                         // Tab -> Next or previous control
                SetFocus (GetNextDlgTabItem(GetParent(hwnd),hwnd,shift));   // Move to next last item.
                return (0);
           case VK_RETURN:                      // Return has special meaning (invoke dialog event)
                SendMessage (GetParent(hwnd),WM_COMMAND,IDOK,0L);
                return (0);
           case VK_ESCAPE:                      // Escape has special meaning (abort dialog)
                SendMessage (GetParent(hwnd),WM_COMMAND,IDCANCEL,0L);
                return (0);
           case VK_HOME:                        // Home -> top of list
                move   (0);
                return (0);
           case VK_END:                         // End -> Bottom of list
                move   (count);
                return (0);
           case VK_UP:                          // Up -> Up one line
                move   (current-1);
                return (0);
           case VK_DOWN:                        // Donw -> Down one line
                move (current+1);
                return (0);
           case VK_PRIOR:                       // Page up -> Up one page
                move (current+1-lines);
                return (0);
           case VK_NEXT:                        // Page down -> Down one page
                move (current+lines-1);
                return (0);
           default:
                break;
         }
         return (0);
//```````````````````````````````````````````````````````````````````
//
//  Scroll bar messages.
//
    case WM_VSCROLL:
         switch (LOWORD(wParam)) {
           case SB_LINEUP:
                i = -1;
                break;
           case SB_LINEDOWN:
                i = 1;
                break;
           case SB_PAGEUP:
                i = 1-lines;
                break;
           case SB_PAGEDOWN:
                i = lines-1;
                break;
           case SB_THUMBTRACK:
           case SB_THUMBPOSITION:
                i = HIWORD(wParam)-top;
                break;
           default:
                return (0);
         }
         j = top;
         top += i;
         if (top > BOTTOM) top = BOTTOM;
         if (top < 0) top = 0;
         if (j == top) return (0);
         scroll ();
         redraw ();
         return (0);
//```````````````````````````````````````````````````````````````````
//
//  Control messages.
//
//  The following are requests for action on the dialog box via messages.
//
    case JL_RESET:                          // Reset contexts of list.
         reset  ();
         return (0);
    case JL_GETJFCLIST:                     // Get the JFC_list object.
         return ((long) this);
//```````````````````````````````````````````````````````````````````
//
//  Mouse events.
//

//
//  Left button up indicates the user is done with the selection.
//
    case WM_LBUTTONUP:
         ReleaseCapture ();
         in_select = false;
         return (0);        
//
//  Mouse moves only count if we are in a mouse select.  If we are, we
//  treat mouse moves simply as if the user shift clicked.
//
    case WM_MOUSEMOVE:
         if (!in_select) return (0);
         wParam = MK_SHIFT;             
//
//  Left mouse button.
//
    case WM_LBUTTONDOWN:            // **** FALL THROUGH ****
         if (msg == WM_LBUTTONDOWN) SetCapture (hwnd);
//
//  Set focus to this window.  Translate messages as needed.
//
         SetFocus (hwnd);
//
//  Find the y location of the mosue click.
//
         i = (HIWORD(lParam)-1)/height+top;
         if (!in_select || (i != last_y)) {             // Mouse selelect that did not really move (suppress flashing).
           last_y    = i;                               // Save last position for next time.
           in_select = true;                            // This could be a mouse select.
//
//  This is a basic left click (or shift-left click
//
           move (i);
         }
//
//  Auto-scroll handler.  When currsor is close enough to the edge 
//  we generate move up or move down commands necessary to scroll the 
//  list.
//
         if ((HIWORD(lParam) < height/2) && (top > 0)) i = SB_LINEUP;
         else if ((HIWORD(lParam) >= height*lines-height/2) && (top < BOTTOM)) i = SB_LINEDOWN;
         else return (0);             // No auto-scroll so exit.

         static short delta = 1;      // This is a KLUDGE used to get around the fact
                                      //   that mouse_event will not generate an event
                                      //   if the mouse does not move so we generate
                                      //   events that move one micky right and left 
                                      //   alternately, so the average is no motion.
         win_proc     (hwnd,WM_VSCROLL,i,0);          // Scroll list.
         UpdateWindow (hwnd);                         // Force window redraw
         mouse_event  (MOUSEEVENTF_MOVE,delta,0,0,0); // Fake mouse event so window keeps scrolling
         if (delta == 1) delta = -1; else delta = 1;  // Toggle mouse direction so no net motion occures.
         return (0);
  }
  return (DefWindowProc(hwnd,msg,wParam,lParam));
}

//
//  End Class JWP_list
//
//-------------------------------------------------------------------

//-------------------------------------------------------------------
//
//  begin class EUC_buffer
//
//  This class implements a simple buffer used to break dictionary 
//  strings into lines, for the display.  
//

//
//  Clear the buffer.  This sets the x poisiton and character count to zero.
//
void EUC_buffer::clear () {
  x     = 0;
  count = 0;
  if (hilight) put_char (EUC_HIGHLIGHT);    // Highlight is generated on the clear command (called 
                                            //   by flush.  This means you need to set the highlight
                                            //   before calling clear.
  return;
}

//
//  Writes the contents of the buffer to the list box control, up to the 
//  specified character.  The remainder of the characters are retained in 
//  the buffer for the next line.
//
//      pos -- Position to flush buffer to.  If this value is -1 the 
//             entire contents ofthe buffer will be flushed.
//
void EUC_buffer::flush (int pos) {
  int  i,j;
  if (pos == -1) pos = count;               // Flush entire buffer.
  i = count;                                // Save lenth of the buffer;
  list->add_line (pos,buffer);              // Write to list-box
  clear    ();                              // Clear buffer.
  put_char ('\t');                          // Indent new line.
  for (j = pos; j < i; j++) put_char (buffer[j]);
  return;
}

//
//  Intialize the buffer system.  This mostly involves caching the list-box
//  window pointer, and determinining the widht of the list box.
//
//      hwnd -- List box window handle.
//
void EUC_buffer::initialize (HWND hwnd) {
  list = (JFC_list *) SendMessage(hwnd,JL_GETJFCLIST,0,0);
  xmax = list->width-sys_font.hwidth;
  highlight (false);
  return;
}

//
//  This is the main routine.  This places a character in the buffer.  
//  If necessary the contents of the buffer are flushed to make room 
//  for the new character.
//
//      ch -- Character to place in the buffer.
//
void EUC_buffer::put_char (int ch) {
  int i;
//
//  If this is a visable (non-space) character and will put us past 
//  the max length marker, then we need to break the line.
//
//  If it is simply a JIS character we can break the line just before 
//  this.  If it is an ascii character, we have to backup until we get 
//  to a space.
//
  if (ch != EUC_HIGHLIGHT) x = sys_font.hadvance(x,ch); // KLUDGE: See top of file, if the character code is 0x01 this is a color shift.
  if (!ISSPACE(ch) && (x >= xmax)) {
    if (ISJIS(ch)) flush (count);
      else {
        for (i = count-1; (i >= 0) && !ISJIS(buffer[i]) && !isspace(buffer[i]); i--);
        i++;
        if (i) flush (i);           // This handles the case when the user has selected a very
      }                             //   large font and a single word of the text will not fit 
  }                                 //   on a line.  This allows the word to extend into the right 
//                                  //   margin.  This is not perfect, but probobly the best.
//  Put the character, converting to EUC code.
//
  if (ISJIS(ch)) ch &= 0x7f7f;
  buffer[count++] = ch;
  return;
}

//
//  Put an entire kanji string into an EUC_buffer.
//
//      kanji  -- Kanji string.
//      length -- Number of characters in the string.
//
void EUC_buffer::put_kanji (KANJI *kanji,int length) {
  int i;
  for (i = 0; i < length; i++) put_char (kanji[i]);
  return;
}

//
//  Puts an entire string into an EUC_buffer.  The string is restricted 
//  to a processing an ascii string.
//
//      string -- String to put into the buffer.
//
void EUC_buffer::put_string (TCHAR *string) {
  while (*string) put_char (*string++);
  return;
}

//
//  End Class EUC_buffer
//
//-------------------------------------------------------------------

//-------------------------------------------------------------------
//
//  Exported routines
//

//
//  This small routine registers the window class used for the japanese
//  edit-box procedure, and Japanese list box procedures.
//
int initialize_edit (WNDCLASS *wclass) {
  wclass->hbrBackground = (HBRUSH) (COLOR_WINDOW+1);
  wclass->style         = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
  wclass->lpfnWndProc   = JFC_text_proc;
  wclass->lpszClassName = TEXT("JFC-Text");
  if (!RegisterClass(wclass)) return (true);
  wclass->lpfnWndProc   = JFC_list_proc;
  wclass->lpszClassName = TEXT("JFC-List");
  if (!RegisterClass(wclass)) return (true);
  return (false);
}









