//-------------------------------------------------------------------//
//                                                                   //
//  JFC Copyright (C) Glenn Rosenthal, 1999,2000,2001.               //
//  All rights reserved.                                             //
//                                                                   //
//-------------------------------------------------------------------//

//-------------------------------------------------------------------
//
//  This module implements japanese versions of the windows controls.
//  In particular, this module implements a japanese text box control
//  This control is used to display most text in this program.
//

#ifndef jfc_edit_h
#define jfc_edit_h

#include "jfc_font.h"

struct JFC_line {           // This structure defines a line break within the display box.
  struct JFC_line *next;    // Pointer to next line break.
  KANJI           *buffer;  // Pointer into buffer for start of this line.
  int              length;  // Number of characters in this line.
};

                            // Mode types used in the mode setting.
#define MODE_SINGLE     0   // Single item display or non-display mode.
#define MODE_MULTI      1   // Multi-line display mode (determined by config).
#define MODE_CYCLE      2   // Ciyclic display mode.
#define MODE_SWAP       3   // Two chanel swap mode.

class JFC_text {
friend class JFC_window;
public:
  int   height;             // Hegith of the text box in pixels

  JFC_text    (void);                                               // Constructor.
  ~JFC_text   (void);                                               // Destructor.
  void adjust (int x,int y);                                        // Adjust window (move without formatting)
  void hide   (void);                                               // Hide the window (used to hide extra windows)
  void move   (int x,int y,int width);                              // Change the location of the window.
  int  open   (int winnum);                                         // Open a window with a specific id number.
  void set    (KANJI *string,int view,class JFC_font *font);        // Set the text and state in the buffer.
  void set    (KANJI *l1,KANJI *l2,KANJI *l3,KANJI *l4,class JFC_font *font);   // Set rotating buffer, or flip buffer.
  void show   (void);                                               // Show this window (see hide)
  void view   (int set);                                            // Determine if text is visisble in window or not.
  int  win_proc (HWND hwnd,int iMsg,WPARAM wParam,LPARAM lParam);   // Window procedure.
  void inline draw (void) { InvalidateRect(window,NULL,true); }
private:
  struct JFC_line *first;           // First display line in list.
  struct JFC_line *last;            // Last display line in list.
  class  JFC_font *font;            // Display font for this character
  KANJI *buffer;                    // Text for this window.
  KANJI *line2,*line3,*line4;       // Line buffers for rotating scroll
  HWND   window;                    // Our Windows window.
  int    width;                     // Width of the window.
  byte   view_state;                // View-state for this window.
  byte   mode;                      // Current view mode.
  byte   multi;                     // Multi-view state.
  byte   active;                    // Indicates the window is active.
  byte   id;                        // Our widnow number.  This determines our ID for messages.
  void   clear         (void);              // Deallocate resources
  void   do_info       (LPARAM lParam);     // Do the processing for character information box.
  int    do_timer      (int key,int menu);  // Process timer sensitive double buttons.
  void   format        (void);              // Format the current display.
  void   format_string (KANJI *string);     // Format a single line into the displa
  int    get_mode      (void);              // Get the current display mode.
  int    new_line      (KANJI *buf);        // Generate a new line.
};

//-------------------------------------------------------------------
//
//  begin class JWP_list
//
//  This class implements a Japanese list box control with lots of 
//  trimmings.
//

//
//  Custom window message values.
//
#define JL_RESET        WMU_RESETLIST           // Clear contents of the list box.
#define JL_GETJFCLIST   WMU_GETWINDOWVALUE      // Get the list structrue

//
//  This class describes a data element from a Japanese list control.
//
typedef class LIST_line {
public:
  void   clear (void);                  // Disposes of all memory associated with this item, and clears flags
  void   alloc (int len,KANJI *line);   // Allocates data for the list, based on an argument.
  short  selected;                      // Non-zero indicates the item is selected.
  short  length;                        // Length of the item.
  KANJI *text;                          // Actuall data.
} LIST_line;


typedef class JFC_list {
public:
  int        count;                     // Number of lines in the list.
  int        width;                     // Width of the list.
  int        current;                   // Current element in the list.
//
//  Constructor/destructor
//
  JFC_list  (HWND hwnd,int l);
  ~JFC_list (void);
//
//  General manipulation commands
//
  void        move          (int pos);                      // Move the cursor.
  void        add_line      (int len,KANJI *text);          // Put a line into the file.
  int         win_proc      (HWND hwnd,int msg,WPARAM wParam,LPARAM lParam);  // Window procedure for the dialog
private: 
  void        draw_line     (HDC hdc,int line);             // Render a specific line
  LIST_line  *get_line      (int line);                     // Find data associated with line
  void        que_line      (int line);                     // Que a line to be redrawn
  void        reset         (void);                         // Clear contents fo the list
  void inline redraw        (void) { InvalidateRect (window,NULL,false); }
  void        select        (int line,int onoff);           // Set selected state or clear it.
  void        scroll        (void);                         // Set scroll bar.

  int         alloc;                // Number of lines allocated for the list.
  int         focus;                // List has the focus
  int         height;               // Height of a line in the list in pixels
  int         in_select;            // Non-zero when user is selecting with the mouse.
  int         last_y;               // Last line selected for mouse selection (to suppress flashing).
  int         lines;                // Visible lines in the display window.
  int         select_count;         // Number of selected lines.
  int         top;                  // Top element in the list.
  KANJI       last_char;            // Character under last mouse click (used for character info).
  HWND        window;               // Window containning the actual list.
  struct LIST_list *lists,*last;    // List and pointer to last list.
} JFC_list;

//
//  End Class JWP_list
//
//-------------------------------------------------------------------

//-------------------------------------------------------------------
//
//  begin class EUC_buffer
//
//  This class implements a simple buffer used to break dictionary 
//  strings into lines, for the display.  
//
#define EUC_HIGHLIGHT   0x01    // Special character for highlight in in list box.  This is
                                //   big KLUDGE, see jwp_edit.cpp for details.  This basically
                                //   indicates that this line is to be in the highlite color.

typedef class EUC_buffer {
public:
  short     x;                                  // Current horizontal position of vertual cursor.
  short     xmax;                               // Maximum acceptable horizontal cursor position.
  HWND      window;                             // Window of list box control to write the string to.
  JFC_list *list;                               // Pointer to JWP_list object
  void  clear      (void);                      // Clears the state of the buffer.
  void  flush      (int pos);                   // Writes the contents of the buffer, up to a specific character.  The rest of the characters are readded to the buffer to form a new line.
  void  inline highlight (int x) {hilight = x;} // Set highlight value.
  void  initialize (HWND hwnd);                 // Initialize the buffer.
  void  put_char   (int ch);                    // Put a character to the buffer.
  void  put_kanji  (KANJI *kanji,int length);   // Put kanji string to buffer.
  void  put_string (TCHAR *text);               // Put a string to the buffer.
private:
  KANJI buffer[256];                            // Buffer to hold the string being built.
  short count;                                  // Current position in the buffer.
  short hilight;                                // Set to non-zero for highlighted entry.
} EUC_buffer;

//
//  End Class EUC_buffer.
//
//-------------------------------------------------------------------

extern int  initialize_edit (WNDCLASS *wclass);         // Intialize the edit class

#endif jfc_edit_h


