//-------------------------------------------------------------------//
//                                                                   //
//  JFC Copyright (C) Glenn Rosenthal, 1999,2000.                    //
//  All rights reserved.                                             //
//                                                                   //
//-------------------------------------------------------------------//

//-------------------------------------------------------------------//
//                                                                   //
//  This module manages the actual card list as a file class object.
//

#include "jfc_main.h"
#include "jfc_conf.h"
#include "jfc_dict.h"
#include "jfc_file.h"
#include "jfc_font.h"
#include "jfc_help.h"
#include "jfc_info.h"
#include "jfc_inst.h"
#include "jfc_jisc.h"
#include "jfc_misc.h"
#include "jfc_stat.h"
#include "jfc_wind.h"

//-------------------------------------------------------------------//
//
//  Exported data
//

class JFC_file *jfc_file;       // Class instance pointer.

//-------------------------------------------------------------------//
//
//  Begin class CARD_data.
//
//  Class used to determine the contents of a card.  This class is only used to parse a 
//  standard card into it's parts.
//

//
//  This routine extracts the data from a normal type card.
//
//      data     -- Data line for card (ie what to parse)
//      encoding -- Data encoding (EUC, UTF-8).
//
//      RETURN   -- A non-zero return voids the card.  This can be used to detect 
//                  errors in the card or comment markers.
//
int CARD_data::parse (byte *data,int encoding) {
  KANJI *ptr,buffer[SIZE_BUFFER*4+1];
  int i,j;
//
//  Convert card from EUC to JIS code array.
//
  if (encoding == ENCODE_EUC) {
    for (i = j = 0; (i < SIZE_BUFFER*4) && data[j]; j++) {
      if (!(data[j] & 0x80)) buffer[i++] = data[j];
        else {
          buffer[i++] = ((data[j] << 8) | data[j+1]) & 0x7f7f;
          j++;
        }
    }
  }
//
//  Decode UTF-8 data.
//
  else {
    for (i = j = 0; (i < SIZE_BUFFER*4) && data[j]; j++) {
      if (UTF8_VALUE2 == (data[j] & UTF8_MASK2)) {
        buffer[i++] = unicode2jis(((data[j] & 0x1f) << 6) | (data[j+1] & 0x3f));
        j++;
      }
      else if (UTF8_VALUE3 == (data[j] & UTF8_MASK3)) {
        buffer[i++] = unicode2jis(((data[j] & 0x0f) << 12) | ((data[j+1] & 0x3f) << 6) | (data[j+2] & 0x3f));
        j += 2;
      }
      else if (UTF8_VALUE4 == (data[j] & UTF8_MASK4)) {
        j += 3;
        buffer[i++] = KANJI_BAD;
      }
      else buffer[i++] = data[j];
    }
  }
  buffer[i] = 0;
//
//  Intiialize the card data structrue.
//
  kanji_known = jfc_config.cfg.know_all_kanji;      // Lets us force all kanji to known
  kanji_only  = false;
  kana_only   = false;
  ascii_only  = false;
  sub_card    = false;
  kanji_card  = false;
//
//  Check for card flags
//
  ptr = buffer;
  while (true) {
    switch (*ptr) {
      case KANJI_HASH:      // This is a comment.
           return (PARSE_SKIP);
      case '&':
      case KANJI_AND:
           for (ptr++; *ptr && ISSPACE(*ptr); ptr++);
           for (i = 0; (i < SIZE_BUFFER) && *ptr; i++) ascii[i] = *ptr++;
           ascii[i] = 0;
           return (PARSE_SUBFILE);
      case 'x':
      case 'X':
      case KANJI_x:
      case KANJI_X:
      case KANJI_CROSS:     // Force kanji is known
           ptr++;
           kanji_known = true;
           break;
      case 'k':
      case KANJI_k:         // Forced kana card
           ptr++;
           kana_only = true;
           break;
      case 'K':
      case KANJI_K:         // Forced kanji card
           ptr++;
           kanji_only = true;
           break;
      case 'a':
      case 'A':
      case KANJI_a:
      case KANJI_A:         // Forced ascii card
           ptr++;
           ascii_only = true;
           break;
      case '+':
      case KANJI_PLUS:      // Kanji sub-card
           ptr++;
           sub_card = true;
           break;
      case '*':
      case KANJI_STAR:      // Kanji card!
           ptr++;
           kanji_card = true;
           break;
      default:
           goto flags_done;
    }
  }
flags_done:
//
//  If this is a kanji card, we interpet the data somewhat differently.  We store 
//  the kanji character in a special spot.  The following mapping for the data fields 
//  is used:
//
//      kanji -- On-yomi
//      kana  -- Kun-yomi
//      ascii -- Mening (imi)
//
  if (kanji_card) {
    while (ISSPACE(*ptr)) ptr++;
    if (!ISKANJI(*ptr)) {
      jfc_window.windows[0].set(ptr,true,&kanji_font);
      jfc_window.adjust (1);
      ErrorMessage (true,IDS_CARD_NOKANJI);
    }
    kanji_char[0] = *ptr;
    kanji_char[1] = 0;
    for (ptr++; ISSPACE(*ptr); ptr++);
//
//  If line contains only kanji (ie after we skip the spaces there is no other text, then
//  use the JWPce kanji inforamtion data to get kanji information for this file.
//
//  This is for non-English users.  If the kanji is followed by ASCII then JFC will insert
//  the on-yomi and the kun-yomi from the kanji inforamtion file, but read the user's 
//  meaning fields.  This allows user to get the readings but replace the meanings.
//
    KANJI_info info;
    if (ISASCII(*ptr) && !info.open_info()) {
      info.get_info   (kanji_char[0],false);
      info.put_data   (ascii,kanji,kana);
      info.close_info ();
      if (*ptr) {                       // Pick up the meaning for cards with user meaning.
        for (i = 0; (i < SIZE_BUFFER) && *ptr; i++) ascii[i] = *ptr++;
        ascii[i] = 0;
      }
      return (PARSE_OK);
    }
//
//  This is not an auto kanji card, the user has supplied some readings for the card. 
//  We could interpret them as a normal card, but it works better with the user to 
//  interpret them separately.  This allows us to process cards without on-yomi correctly.
//
    static KANJI no_value[] = { '-','n','o',' ','r','e','a','d','i','n','g','s','-',0 };
    kstrcpy (kanji,no_value);
    kstrcpy (kana ,no_value);
    for (i = j = 0; ISKANA(*ptr); ) {   // While there are readings
      if (ISKATAKANA(*ptr)) {
        if (i) kanji[i++] = KANJI_COMMA;
        for (; (i < SIZE_BUFFER) && *ptr && !ISSPACE(*ptr); i++) kanji[i] = *ptr++;
        kanji[i] = 0;
        while (ISSPACE(*ptr)) ptr++;
      }
      else {
        if (j) kana[j++] = KANJI_COMMA;
        for (; (j < SIZE_BUFFER) && *ptr && !ISSPACE(*ptr); j++) kana[j] = *ptr++;
        kana[j] = 0;
        while (ISSPACE(*ptr)) ptr++;
      }
    }					                // Everything else is meaning
    for (i = 0; (i < SIZE_BUFFER) && *ptr; i++) ascii[i] = *ptr++;
    ascii[i] = 0;
    return (PARSE_OK);
  }
//..................................................................................
//
//  Vocab cards.
//
//  Find first space.  To start with we will assume this is the kanji data.
//
  while (ISSPACE(*ptr)) ptr++;
  for (i = 0; (i < SIZE_BUFFER) && *ptr && !ISSPACE(*ptr); i++) kanji[i] = *ptr++;
  kanji[i] = 0;
  while (ISSPACE(*ptr)) ptr++;
//
//  If we just have one word, then this is really just an entry to look up in the 
//  dictionary so here we go.
//
  KANJI result[SIZE_BUFFER];
  if (!*ptr) {
    JFC_dict dict;
    if (!dict.search_dict(kanji,result)) {
      jfc_window.windows[0].set(kanji,true,&kana_font);
      jfc_window.adjust (1);
      ErrorMessage (false,IDS_CARD_NODICT);
    }
    else {
      encoding = ENCODE_EUC;
      ptr      = result;
      for (i = 0; (i < SIZE_BUFFER) && *ptr && !ISSPACE(*ptr); i++) kanji[i] = *ptr++;
      kanji[i] = 0;
      while (ISSPACE(*ptr)) ptr++;
    }
  }
//
//  Check next string, if this beginning with a JIS character assume this is a kana string.
//
  if ((*ptr == '[') || (*ptr == KANJI_LBRACKET)) ptr++;             // Skip [ for JWPce dictionary entry
  if (ISJIS(*ptr)) {                                                // Was JIS so have kanji and kana.
    for (i = 0; (i < SIZE_BUFFER) && *ptr && !ISSPACE(*ptr); i++) kana[i] = *ptr++;
    if ((kana[i-1] == ']') || (kana[i-1] == KANJI_RBRACKET)) i--;   // Skip ] for JWPce dictionary entry
    kana[i] = 0;
  }
  else {                                                        // Just have a kana string.
    memcpy (kana,kanji,sizeof(kana));
    kanji[0] = 0;
  }
//
//  The rest must be ascii data.
//
  while (ISSPACE(*ptr)) ptr++;
  for (i = 0; (i < SIZE_BUFFER) && *ptr; i++) ascii[i] = *ptr++;
  ascii[i] = 0;
  return (PARSE_OK);
}

//
//  Set the main window display for working with a standard card.
//
void CARD_data::set () {
  if (kanji[0]) {					// Card with kanji
    jfc_window.windows[0].set (kanji,false,&kanji_font);
    jfc_window.windows[1].set (kana ,false,&kana_font );
    jfc_window.windows[2].set (ascii,false,&imi_font  );
    jfc_window.adjust (3);
  }
  else {						// Card without kanji
    jfc_window.windows[0].set (kana ,false,&kana_font );
    jfc_window.windows[1].set (ascii,false,&imi_font  );
    jfc_window.adjust (2);
  }
  return;
}

//                                                                   //
//  End class CARD_data.
//
//-------------------------------------------------------------------//

//-------------------------------------------------------------------//
//                                                                   //
//  Data structures for statisitcs and save files.
//
//  There are two versions of the stat files.  The old version is at MAGIC_FILE-1, and the 
//  new version is at MAGIC_FILE.  The old version did not correctly save or load review
//  settings.  There was space for these is in the file but the values were not correctly
//  written.  All files are automatically written with the new settings.  When reading
//  older files the review mode settings remain unchanged.
//
//  The format of the save file is a follows:
//
//      FileSaveHeader      -- General header.  
//      { FileSubFile }     -- An array of FileSubFile structures.  The header->files
//                             field indicates the number of these.  Each of these structures
//                             is used to check the date and size of each subfile.  Remember
//                             the main file is always loaded as the first subfile.
//      { FileFlash }       -- The reaminning flash cards.  The number of these is indicated
//                             byte the header->count field.  Each record represents a single
//                             card.  The order matches the order of the cards.
//

#define MAGIC_FILE  0x4cd739eb      // Magic ID to make sure this is a statisitics file.
#define MAGIC_SAVE  0xa43094c0      // Magic ID for saved files.

struct FileCardInfo {               // Structure for each card
  byte stat_kanji;
  byte stat_kana;
  byte stat_ascii;
  byte fill;                        // Not used.
};

struct FileHeader {                 // File structure for stat file.
  long     size;                    // Size of main file.
  FILETIME time;                    // Last write time for main file.
  long     magic;                   // Identifies this as a JFC stat file.
  long     count;                   // Number of cards.
  byte     flash_kanji;             // Show kanji and let user guess kana & meaning
  byte     flash_kana;              // Show kana and let user guess kanji & meaning
  byte     flash_ascii;             // Show ascii and let user guess kana & kanji
  byte     kanji_kanji;             // Show kanji and let user guess readings & meaning (kanji cards)
  byte     kanji_kana;              // Show readings and let user guess kanji & meaning (kanji cards)
  byte     kanji_ascii;             // Show ascii and let user guess readings & kanji (kanji cards)
  byte     kanji_always;            // Always show kanji with kana.
  byte     kanji_unknown;           // Only show kanji witht he kana when it is unknown
  byte     duplicate_cards;         // Duplicate cards that are missed.
  byte     knownkanji_colorkanji;   // Assume that color kanji indicates the known kanji.
  byte     review_mode;             // Use review mode and the weights.
  byte     flashkanji[9];           // Flash kanji weights.
  byte     flashkana [9];           // Flash kana weights.
  byte     flashascii[9];           // Flash ascii weights.
  byte     kanjikanji[9];           // Kanji card kanji weight
  byte     kanjikana [9];           // Kanji card kana weight
  byte     kanjiascii[9];           // Kanji card ascii weight
  short    duplicate_count;         // Frequency to duplicate cards 
  byte     know_all_kanji;          // All kanji are known
  byte     no_move_end;             // Do not place wrong card at the end.
  byte     time_cards;              // If user does not answer in time the card is wrong.
  byte     use_max;                 // Limit number of cards generated.
  int      maximum;                 // Maximum number of cards generated.
  int      time_show;               // Time to show the card in msec.
  int      time_next;               // Time to move to the next card in msec.
  byte     fill[48];                // Expansion for later.
  struct FileCardInfo cards;        // The first element in the card data.
};

struct FileSaveHeader {             // Header used for save files.
  long     magic;                   // Identifies this as a JFC stat file.
  long     count;                   // Number of cards.
  long     right;                   // Number of right cards.
  long     wrong;                   // Number of wrong cards.
  long     files;                   // Number of sub-files contained in the file.
};

struct FileSubFile {                // Record used for storing information about subfiles
  long     size;                    // Size of the sub-file
  FILETIME time;                    // Last write time for main file.
};

struct FileFlash {                  // Record sturcture for saved cards
  int  card;                        // Index to card data (zero based)
  byte duplicate;                   // Duplicate card flag
  byte type;                        // Card type
  byte prob;                        // Probability (not really used).
  byte fill;                        // Not used.
};

//-------------------------------------------------------------------//
//
//  Begin class SUB_file.
//
//  This class manages a sub-file and keeps track of all the cards in it.
//  All the sub-files are merged together in the JFC_file class.
//

//
//  Add a card the curretn sub file.  Later the varius lists from the different sub-files
//  will be merged later.
//
//      dat    -- Data for card.
//
//      RETURN -- A non-zero value indicates an error.
//
int SUB_file::add_card (byte *dat) {
  int i;
  class JFC_card *card;
//
//  Skip spaces at the beginning of the line.  Abort the card if the line is blank or the 
//  card has a comment marker.
//
  while (ISSPACE(*dat)) dat++;
  if (!*dat || (*dat == '#')) return (false);
//
//  Decode the data so we can check the first character in the card
//
  class CARD_data card_data;
  switch (card_data.parse(dat,encoding)) {
    case PARSE_OK:                              // Process the card
         break;
    case PARSE_SKIP:                            // This picks up errors and # comments
         return (false);
    case PARSE_SUBFILE: {                       // The sub-files.
           TCHAR temp[SIZE_BUFFER],buffer[SIZE_BUFFER];
#ifdef WINCE
           for (i = 0; card_data.ascii[i]; i++) temp[i] = jis2unicode(card_data.ascii[i]);
           temp[i] = 0;
           if (temp[0] == '\\') lstrcpy (buffer,temp);
             else {
               lstrcpy (buffer,filename);
               for (i = lstrlen(buffer); i && (buffer[i] != '\\'); i--);
               if (buffer[i] == '\\') buffer[i+1] = 0;
               lstrcat (buffer,temp);
             }
#else  WINCE
           for (i = 0; card_data.ascii[i]; i++) temp[i] = (byte) card_data.ascii[i];
           temp[i] = 0;
           if (temp[1] == ':') lstrcpy (buffer,temp);
             else {
               lstrcpy (buffer,filename);
               for (i = lstrlen(buffer); i && (buffer[i] != '\\'); i--);
               if (buffer[i] == '\\') buffer[i+1] = 0;
               lstrcat (buffer,temp);
             }
#endif WINCE
           jfc_file->read_subfile (buffer);
         }
         return (false);
  }
//
//  Link the card into the data.
//
  if (!(card = new class JFC_card)) return (true);      // Create card structure and fill it out.
  if (last) last->next = card; else cards = card;
  last             = card;
  card->next       = NULL;
  card->data       = dat;
  card->flags      = 0;
  card->stat_ascii = 0;
  card->stat_kana  = 0;
  card->stat_kanji = 0;
  card->encoding   = encoding;
  count++;
//
//  Decode the card so we can set the card flags.
//
  if (card_data.ascii_only) card->flags |= CARD_ASCIIONLY;
  if (card_data.kanji_card) card->flags |= CARD_KANJI;
  if (card_data.sub_card  ) card->flags |= CARD_SUBCARD;
  if (card_data.kana_only ) card->flags |= CARD_KANAONLY;
  if (card_data.kanji[0]) {
    if (card_data.kanji_only) 
    card->flags |= CARD_KANJIONLY;
    card->flags |= CARD_HASKANJI;
    if      (card_data.kanji_known) card->flags |= CARD_USEKANJI;
    else if (jfc_config.cfg.knownkanji_colorkanji) {
      for (i = 0; card_data.kanji[i]; i++) {
        if (ISKANJI(card_data.kanji[i]) && !in_colorkanji(card_data.kanji[i])) break;
      }
      if (!card_data.kanji[i]) card->flags |= CARD_USEKANJI;
    }
  }
  return (false);
}

//
//  This routine loads a specifc subfile.  
//
//      name   -- Name of file to load.
//      config -- If non-zero allows the configuration to be read from the statistics
//                file.  For this to occure, however, the stting in JFC_config must still
//                be set to read properties from the config file.  This is normally only
//                set for the base file (the one the user opend).  For all other sub-files
//                this should be false.
//
//      RETURN -- Non-zero value indicates an error.
//
int SUB_file::load (TCHAR *name,int config) {
  HANDLE file;
  long i,j;
  BY_HANDLE_FILE_INFORMATION info;
//
//  Setup file name
//
  if (!(filename = strdup(name))) return (true);
//
//  Actually read the file.  Take apart each line and convert it into card data.
//
  if (INVALID_HANDLE_VALUE == (file = OPENREAD(name))) return (true);
  GetFileInformationByHandle (file,&info);
  size = info.nFileSizeLow;
  time = info.ftLastWriteTime;
  if (!(buffer = (byte *) calloc(size+4,1))) { CloseHandle (file); return (true); }
  ReadFile    (file,buffer,size,(unsigned long *) &i,NULL);
  CloseHandle (file);
//
//  Determine the file encoding.
//
  encoding = ENCODE_EUC;
  if (is_utf8(buffer,size)) encoding = ENCODE_UTF8;
//
//  Extract the cards from the file.
//
  for (i = 0; i < size; i++) {
    if (ISCRLF(buffer[i])) continue;
    for (j = i; (i < size) && !ISCRLF(buffer[i]); i++);
    buffer[i] = 0;
    if (add_card(buffer+j)) return (true);
  }
//..........................................................................
//
//  Attempt to read the statistics file.
//
//  Open the file and read the contents into a memory buffer.
//
  struct FileHeader *temp;
  class  JFC_card   *card;
  TCHAR  stat[SIZE_BUFFER];

  if (INVALID_HANDLE_VALUE == (file = OPENREAD(stat_name(stat)))) return (false);
  i = GetFileSize(file,NULL);
  if (!(temp = (struct FileHeader *) calloc(i+4,1))) { CloseHandle (file); return (true); }
  ReadFile (file,temp,i,(unsigned long *) &i,NULL);
  CloseHandle (file);
  if ((temp->magic != MAGIC_FILE) && (temp->magic != MAGIC_FILE-1)) {
    ErrorMessage (true,IDS_FILE_NOTSTAT,stat);
    free         (temp);
    return       (true);
  }
//
//  Check and make sure this statistic file matches the card file.  If not give the user the 
//  choice of using it anyway.
//
  if (CompareFileTime(&temp->time,&time) || (temp->size != size)) {
    if (IDYES != JMessageBox (main_window,IDS_STAT_TEXT,IDS_STAT_TITLE,MB_YESNO | MB_ICONWARNING | MB_DEFBUTTON2,stat)) { free (temp); return (false); }
  }
//
//  Read the card statistics.
//
  if (temp->count > count) temp->count = count;                 // Make sure the count is OK.
  for (i = 0, card = cards; i < temp->count; i++, card = card->next) {
    card->stat_kanji = (&temp->cards)[i].stat_kanji;
    card->stat_kana  = (&temp->cards)[i].stat_kana;
    card->stat_ascii = (&temp->cards)[i].stat_ascii;
  }
//
//  Read the configuration from the statistics file.
//
  if (config && jfc_config.cfg.stat_config) {
    jfc_config.cfg.flash_kanji           = temp->flash_kanji;
    jfc_config.cfg.flash_kana            = temp->flash_kana;
    jfc_config.cfg.flash_ascii           = temp->flash_ascii;
    jfc_config.cfg.kanji_kanji           = temp->kanji_kanji;
    jfc_config.cfg.kanji_kana            = temp->kanji_kana;
    jfc_config.cfg.kanji_ascii           = temp->kanji_ascii;
    jfc_config.cfg.kanji_always          = temp->kanji_always;
    jfc_config.cfg.kanji_unknown         = temp->kanji_unknown;
    jfc_config.cfg.know_all_kanji        = temp->know_all_kanji;
    jfc_config.cfg.duplicate_cards       = temp->duplicate_cards;
    jfc_config.cfg.duplicate_count       = temp->duplicate_count;
    jfc_config.cfg.knownkanji_colorkanji = temp->knownkanji_colorkanji;
    jfc_config.cfg.review_mode           = temp->review_mode;
    jfc_config.cfg.no_move_end           = temp->no_move_end;
    jfc_config.cfg.time_cards            = temp->time_cards;
    jfc_config.cfg.use_max               = temp->use_max;
    jfc_config.cfg.maximum               = temp->maximum;
    jfc_config.cfg.time_show             = temp->time_show;
    jfc_config.cfg.time_next             = temp->time_next;
    if (jfc_config.cfg.duplicate_count <= 4) jfc_config.cfg.duplicate_count = 5;
//
//  Read review mode settings.
//
//  There are two versions of the stat files.  The old version is at MAGIC_FILE-1, and the 
//  new version is at MAGIC_FILE.  The old version did not correctly save or load review
//  settings.  There was space for these is in the file but the values were not correctly
//  written.  All files are automatically written with the new settings.  When reading
//  older files the review mode settings remain unchanged.
//  
    if (temp->magic == MAGIC_FILE) {
      for (i = 0; i < 9; i++) {
        jfc_config.cfg.flashkanji[i] = temp->flashkanji[i];
        jfc_config.cfg.flashkana [i] = temp->flashkana [i];
        jfc_config.cfg.flashascii[i] = temp->flashascii[i];
        jfc_config.cfg.kanjikanji[i] = temp->kanjikanji[i];
        jfc_config.cfg.kanjikana [i] = temp->kanjikana [i];
        jfc_config.cfg.kanjiascii[i] = temp->kanjiascii[i];
      }
    }
  }
//
//  Deallocate the resources.
//
  free   (temp);
  return (false);
}

//
//  Save the statistics file.
//
//      RETURN -- A non-zero value indicates an error.
//
int SUB_file::save_stats () {
  TCHAR  name[SIZE_BUFFER];
  HANDLE file;
  class JFC_card *card;
  struct FileHeader *temp;
  int    i,s;
//
//  Allocate buffer for the data.  This is a memory image of the file.  Also try to open the 
//  file.
//
  s = sizeof(struct FileHeader)+sizeof(struct FileCardInfo)*count;
  if (INVALID_HANDLE_VALUE == (file = OPENWRITE(stat_name(name)))) return (true);
  if (!(temp = (struct FileHeader *) calloc(s,1))) { CloseHandle (file); return (true); }
//
//  Write the configuration to the buffer.
//
  temp->size                  = size;               // Size of card file for checking.
  temp->time                  = time;               // Time of card file for checking.      
  temp->magic                 = MAGIC_FILE;         // This is a statistic file.
  temp->count                 = count;              // Number of cards in the file.
  temp->flash_kanji           = jfc_config.cfg.flash_kanji;
  temp->flash_kana            = jfc_config.cfg.flash_kana;
  temp->flash_ascii           = jfc_config.cfg.flash_ascii;
  temp->kanji_kanji           = jfc_config.cfg.kanji_kanji;
  temp->kanji_kana            = jfc_config.cfg.kanji_kana;
  temp->kanji_ascii           = jfc_config.cfg.kanji_ascii;
  temp->kanji_always          = jfc_config.cfg.kanji_always;
  temp->kanji_unknown         = jfc_config.cfg.kanji_unknown;
  temp->know_all_kanji        = jfc_config.cfg.know_all_kanji;
  temp->duplicate_cards       = jfc_config.cfg.duplicate_cards;
  temp->duplicate_count       = jfc_config.cfg.duplicate_count;
  temp->knownkanji_colorkanji = jfc_config.cfg.knownkanji_colorkanji;
  temp->review_mode           = jfc_config.cfg.review_mode;
  temp->no_move_end           = jfc_config.cfg.no_move_end;
  temp->time_cards            = jfc_config.cfg.time_cards;
  temp->use_max               = jfc_config.cfg.use_max;
  temp->maximum               = jfc_config.cfg.maximum;
  temp->time_show             = jfc_config.cfg.time_show;
  temp->time_next             = jfc_config.cfg.time_next;

  for (i = 0; i < 9; i++) {
    temp->flashkanji[i] = jfc_config.cfg.flashkanji[i];
    temp->flashkana [i] = jfc_config.cfg.flashkana [i];
    temp->flashascii[i] = jfc_config.cfg.flashascii[i];
    temp->kanjikanji[i] = jfc_config.cfg.kanjikanji[i];
    temp->kanjikana [i] = jfc_config.cfg.kanjikana [i];
    temp->kanjiascii[i] = jfc_config.cfg.kanjiascii[i];
  }
//
//  Write the card data to the memory image.
//
  for (i = 0, card = cards; i < count; i++, card = card->next) {
    (&temp->cards)[i].stat_kanji = card->stat_kanji;
    (&temp->cards)[i].stat_kana  = card->stat_kana;
    (&temp->cards)[i].stat_ascii = card->stat_ascii;
  }
//
//  Do the write and finish up.
//
  WriteFile (file,temp,s,(unsigned long *) &i,NULL);
  free (temp);
  CloseHandle (file);
  return (false);
}

//
//  This routine builds the stat-file name for the current file.  This is basically the main 
//  file name with the extension .jst.
//
//      buffer -- Buffer to build the file name in (assumed to be of adequate size).
//      
//      RETURN -- Pointer to buffer (just for convenience).
//
TCHAR *SUB_file::stat_name (TCHAR *buffer) {
  int i;
  lstrcpy (buffer,filename);
  for (i = lstrlen(buffer)-1; (i >= 0) && (buffer[i] != ':') && (buffer[i] != '//') && (buffer[i] != '.'); i--);
  if (buffer[i] == '.') buffer[i] = 0;
  lstrcat (buffer,TEXT(".jst"));
  return (buffer);
}

//
//  End class SUB_file.
//
//-------------------------------------------------------------------//

//-------------------------------------------------------------------//
//
//  Begin class JFC_file.
//
//  This class manages a file, and thus an entire deck of cards.
//

//
//  Constructor
//
JFC_file::JFC_file () {
  memset (this,0,sizeof(JFC_file));
  return;
}

//
//  Destructor.
//
JFC_file::~JFC_file () {
  class JFC_card  *card;
  class SUB_file  *file;
  if (filename) free (filename);    // Delete file name.
  while (cards) {                   // Delete master cards set.
    card  = cards;
    cards = card->next;
    delete card;
  }
  while (flash) flash_delete ();    // Delete flash cards.
  while (files) {
    file        = files;
    files       = file->next;
    file->cards = NULL;
    delete file;
  }
  memset (this,0,sizeof(JFC_file));
  return;
}

//
//  Closes the current file.  All resource are deallocated.
//
//      check  -- If set to non-zero the user will be given the chance to abort the file close.
//
//      RETURN -- A non-zero value indicates the user canceled the close.
//
int JFC_file::close (int check) {
  if (!this) return (false);
//
//  Handle the checking.
//
  if (flash && check) {
    switch (jfc_config.cfg.exit_open_file) {
      case EXIT_ASK:
           switch (ButtonDialog(IDD_SAVECHECK,filename,IDH_FILE_SAVE)) {
             case IDOK: 
                  if (save()) return (true);
                  break;
             case IDC_SCNO:
                  break;
             case IDCANCEL:
                  return (true);
           }
           break;
      case EXIT_SAVE:
           if (save()) return (true);
           break;
      case EXIT_QUIT:
           break;
    }
  }
//
//  Saving the stats.
//  
  if (jfc_config.cfg.keep_stats) {
    class SUB_file *file;
    for (file = files; file; file = file->next) {
      if (file->save_stats()) {
        TCHAR name[SIZE_BUFFER];
        ErrorMessage (true,IDS_FILE_STATSAVE,file->stat_name(name));
      }
    }
  }
//
//  Do the actual close.
//
  delete (this);
  jfc_file = NULL;
  jfc_window.adjust (0);
  window_title      ();
  EnableMenuItem    (hmenu,IDM_FILE_CLOSE ,MF_BYCOMMAND | MF_GRAYED);
  EnableMenuItem    (hmenu,IDM_FILE_REOPEN,MF_BYCOMMAND | MF_GRAYED);
  return (false);
}

//
//  Process menu commands (and all others).  This is the main command processor.
//
//      wParam -- The wParam command from the message.  This is basically the menu command ID.
//
void JFC_file::do_menu (int wParam) {
  int i;
  switch (LOWORD(wParam)) {
// --------------------------------------------- File menu
    case IDM_FILE_OPEN:
         if (close (true)) break;
         open_file (NULL);
         break;
    case IDM_FILE_REOPEN:
         if (this && YesNo(IDS_FILE_REOPEN)) reinit (false);
         break;
    case IDM_FILE_CLOSE:
         close (true);
         break;
    case IDM_FILE_DELETE:
         if (this && filename) {
           TCHAR stat[SIZE_BUFFER],save[SIZE_BUFFER];
           if (YesNo(IDS_FILE_DELETE,filename)) {
             DeleteFile (filename);         // Delete the flash card file.
             stat_name  (stat);             // Get the stat file name
             save_name  (save);             // Get the save file name.
             close      (false);            // Close the file (this will write the stat-file
             DeleteFile (stat);             // Delete the stat file.
             DeleteFile (save);             // Delete the save file.
           }
         }
         break;
    case IDM_FILE_EXIT:
         jfc_config.write_files ();
         if (!close(true)) DestroyWindow (main_window);
         break;
// --------------------------------------------- Card menu
    case IDM_CARD_SHOW:
         if (jfc_config.cfg.time_cards) {
           KillTimer (main_window,TIME_SHOW);
           if (jfc_config.cfg.time_next) SetTimer  (main_window,TIME_NEXT,jfc_config.cfg.time_next,NULL);
         }
         if (!flash) break;
         for (i = 0; i < jfc_window.number_windows; i++) jfc_window.windows[i].view (true);
         break;
    case IDM_CARD_RIGHT:
         if (jfc_config.cfg.time_cards) {
           KillTimer (main_window,TIME_SHOW);
           KillTimer (main_window,TIME_NEXT);
         }
         if (!flash) break;
         do_stats     (true);
         flash_delete ();
         if (flash_done()) break;
         flash_next ();
         break;
    case IDM_CARD_WRONG:
         if (jfc_config.cfg.time_cards) {
           KillTimer (main_window,TIME_SHOW);
           KillTimer (main_window,TIME_NEXT);
         }
         if (!flash) break;
         do_stats (false);
         if (flash->duplicate == DUPLICATE_DUPLICATE) {     // Handle a missed duplicate card
           flash_delete ();
           flash_next   ();
           break;
         }
         if (flash->next) {
           JFC_flash *flsh,*f;
           if (jfc_config.cfg.duplicate_cards) {            // Create duplicates
             flash->duplicate = DUPLICATE_DUPLICATE;
             flsh = flash;
             while (true) {
               for (i = 0; flsh && flsh->next && (i < jfc_config.cfg.duplicate_count); i++) flsh = flsh->next;
               if (!flsh || !flsh->next) break;
               if (f = new JFC_flash) {
                 *f         = *flash;
                 f->next    = flsh->next;
                 flsh->next = f;
                 flash_count++;
               }
             }
           }
           if (jfc_config.cfg.no_move_end) flash_delete (); // If we are 
             else {                                         // Push original card to end
               flash->duplicate = DUPLICATE_ORIGINAL;         
               for (flsh = flash; flsh->next; flsh = flsh->next);
               f          = flash;
               flsh->next = flash;
               flash      = flash->next;
               f->next    = NULL;
             }
         }
         if (flash_done()) break;
         flash_next ();
         break; 
// --------------------------------------------- Utilities menu
    case IDM_UTILITIES_OPTIONS:
         jfc_config.options ();
         break;
    case IDM_UTILITIES_DEFAULTOPTIONS:
         jfc_config.def ();
         break;
    case IDM_UTILITIES_SAVESETTINGS:
         jfc_config.write ();
         break;
    case IDM_UTILITIES_INSTALL:
         do_install (true);
         break;
// --------------------------------------------- Recent Files
    case IDM_HELP_MAININDEX:
         do_help (main_window,0);
         break;
    case IDM_HELP_ABOUT: {
           TCHAR text[SIZE_BUFFER],title[SIZE_BUFFER];
           format_string (text ,IDS_ABOUT_TEXT ,VERSION_STRING);
           format_string (title,IDS_ABOUT_TITLE,VERSION_STRING);
           MessageBox (main_window,text,title,MB_OK);
         }
         break;
// --------------------------------------------- Recent Files
    default:
         i = LOWORD(wParam);
         if ((i >= IDM_FILE_FILES_BASE) && (i <= IDM_FILE_FILES_BASE+9)) {
           TCHAR buffer[SIZE_BUFFER];
           get_menudata (hmenu,i,false,buffer);
           if (!close(true)) open_file (buffer+3);
         }
         break;
  }
  return;
}

//
//  This rotuine is called whenever a card is correct or incorrect to update the cards 
//  statisitcs.  This include modifiy the overall stat couns for going through the deck,
//  and the card count for dealing with each card.
//
//      ok -- Set to true if the card was correct and false if the card was incorect.
//
//
void JFC_file::do_stats (int ok) {
  JFC_card *card = flash->card;
  byte *stat;
  if (flash->duplicate) return;             // Skip duplicate cards.
  if (ok) right++; else wrong++;            // Change major stats.
  switch (flash->type) {                    // Get status byte for card.
    case FLASH_KANJI:
    case FLASH_KANJIKANJI:
         stat = &card->stat_kanji;
         break;
    case FLASH_KANA:
    case FLASH_KANJIKANA:
         stat = &card->stat_kana;
         break;
    case FLASH_ASCII:
    case FLASH_KANJIASCII:
         stat = &card->stat_ascii;          
         break;
  }
  *stat = ((*stat) << 1) | (ok & 0x01);     // Modify status byte
  return;
}

//
//  Adds a flash card to the beiginning of the list.
//
//      _card  -- Pointer to the actual card data.
//      _type  -- Indicates the type of card to generate.
//
//      RETURN -- A non-zero value indicates an error.
//
int JFC_file::flash_add (class JFC_card *_card,int _type) {
  JFC_flash *flsh;
  byte prob;
//
//  If we are in review mode then we may not need to add this card.
//
  if (!jfc_config.cfg.review_mode) prob = 100;      // No stats so cards all get 100% chance.
    else {
      byte temp,stat,*array;
      int  i,j;
      switch (_type) {                              // Determine which stats to use.
        case FLASH_KANJIKANJI:
             stat  = _card->stat_kanji;
             array = jfc_config.cfg.kanjikanji;
             break;
        case FLASH_KANJIKANA:
             stat  = _card->stat_kana;
             array = jfc_config.cfg.kanjikana;
             break;
        case FLASH_KANJIASCII:
             stat  = _card->stat_ascii;
             array = jfc_config.cfg.kanjikana;
             break;
        case FLASH_KANJI:
             stat  = _card->stat_kanji;
             array = jfc_config.cfg.flashkanji;
             break;
        case FLASH_KANA:
             stat  = _card->stat_kana;
             array = jfc_config.cfg.flashkana;
             break;
        case FLASH_ASCII:
             stat  = _card->stat_ascii;
             array = jfc_config.cfg.flashascii;
             break;
        case FLASH_ERROR:                           // Special error type, that cannot be suppressed.
             stat  = 0xff;
             temp  = 100;
             array = &temp;
             break;
      }
      for (j = i = 0; i < 8; i++) {                 // Count the number of corrects.
        j += (stat & 0x01);
        stat >>= 1;
      }
      prob = array[8-j];
      if (prob < 1+random(100)) return (false);     // Do we want this card?
    }
//
//  Actually make a card.
//
  if (!(flsh = new JFC_flash)) return (true);
  flsh->next      = flash;
  flsh->card      = _card;
  flsh->type      = (_type == FLASH_ERROR) ? FLASH_ASCII : _type;
  flsh->duplicate = false;
  flsh->prob      = prob;
  flash = flsh;
  flash_count++;
  return (false);
}

//
//  Delete the topmost flash card.
//
void JFC_file::flash_delete () {
  JFC_flash *flsh;
  if (!flash) return;
  flsh  = flash;
  flash = flsh->next;
  delete flsh;
  flash_count--;
  return;
}

//
//  Determins if a file is done or not.  If the cards are done, this routine will do
//  all the necessary cleanup, including the summary screen and all.
//
//      RETURN -- A non-zero return value indicates the set is done.
//
int JFC_file::flash_done () {
  if (flash) return (false);
  jfc_stat.redraw ();
  JMessageBox (main_window,IDS_DONE_TEXT,IDS_DONE_TITLE,MB_OK | MB_ICONINFORMATION,right,(right*100)/(right+wrong),wrong,(wrong*100)/(right+wrong));
  close (false);
  jfc_config.write_files ();
#ifdef xWINCE_PPC                   // This was an option to auto-go to the next file.  
  SendMessage (main_window,WM_COMMAND,IDM_FILE_OPEN,0);
#endif xWINCE_PPC
  return (true);
}

//
//  This routine generates the display for all kanji-type flash cards.
//
void JFC_file::flash_kanji () {
  JFC_card *card,*sub;
  int count,i,j,k,l;
  byte array[101];
  class CARD_data card_data,sub_card;
//
//  Process the actual kanji card.
//
  card = flash->card;
  card_data.parse (card->data,card->encoding);
//
//  If this is an ascii or reading version of kanji card, drop all sub-cards.
//
  if ((flash->type == FLASH_KANJIKANA) || (flash->type == FLASH_KANJIASCII)) {
    count = 0;
  }
//
//  Kanji version of card so, Process all simple cards.
//
  else {
    for (count = 0, sub = card->next; sub; sub = sub->next, count++) {
      sub_card.parse (sub->data,card->encoding);
      if (!sub_card.sub_card) break;
    }
  }
//
//  Display the basic card.
//
  if (!count || jfc_config.cfg.full_kanjicard) {
    jfc_window.windows[0].set (card_data.kanji_char,flash->type == FLASH_KANJIKANJI,&kanji_font);
    jfc_window.windows[1].set (card_data.kanji     ,flash->type == FLASH_KANJIKANA ,&kana_font );
    jfc_window.windows[2].set (card_data.kana      ,flash->type == FLASH_KANJIKANA ,&kana_font );
    jfc_window.windows[3].set (card_data.ascii     ,flash->type == FLASH_KANJIASCII,&imi_font  );
    i = 4;
  }
  else {
    jfc_window.windows[0].set (card_data.kanji_char,card_data.kanji,card_data.kana,card_data.ascii,&kanji_font);
    i = 1;
  }
//
//  Build random array for sub-cards.
//
  for (j = 0; j < 100; j++) array[j] = j;
  if (count > 100) count = 100;
//
//  Generate eah of the sub-cards.
//
  while ((count > 0) && (i < NUMBER_WINDOWS)) {
    j = random(count);
    k = array[j];
    for (l = j+1; l < count; l++) array[l-1] = array[l];
    count--;
    for (sub = card->next; k > 0; k--) sub = sub->next;
    sub_card.parse (sub->data,sub->encoding);
    jfc_window.windows[i++].set (sub_card.kanji,sub_card.kana,sub_card.ascii,NULL,jfc_config.cfg.kana_for_advanced ? &kana_font : &kanji_font);
  }
//
//  Do the actual display.
//
  jfc_window.adjust (i);
  jfc_window.set    (1,false);          // Set the first window as active.
  return;
}

//
//  This routine displays the current flash card on the top of the stack for the user.
//
void JFC_file::flash_next () {
  JFC_card *card;
  class CARD_data card_data;

  if (!flash) return;
  jfc_stat.redraw ();
//
//  Process kanji cards specially.
//
  if ((flash->type == FLASH_KANJIKANJI) || (flash->type == FLASH_KANJIKANA) || (flash->type == FLASH_KANJIASCII)) {
    flash_kanji ();
    if (jfc_config.cfg.time_cards && jfc_config.cfg.time_show) SetTimer (main_window,TIME_SHOW,jfc_config.cfg.time_show,NULL);
    return;
  }
//
//  Process all simple cards.
//
  card = flash->card;                           // Get the card.
  card_data.parse (card->data,card->encoding);  // Extract the data 
  card_data.set   ();                           // Set the data fields.
  switch (flash->type) {                        // Special information for dealing with each card type.
    case FLASH_KANJI:                           // Show the kanji 
         jfc_window.windows[0].view (true);
         break;
    case FLASH_ASCII:                           // Show the ASCII data.
         jfc_window.windows[(card->flags & CARD_HASKANJI) ? 2 : 1].view (true);
         break;
    case FLASH_KANA:                            // Show the kana data.  May need to show the kanji data too.
         if (!(card->flags & CARD_HASKANJI)) jfc_window.windows[0].view (true);
           else {
             jfc_window.windows[1].view (true);
             if (jfc_config.cfg.kanji_always || (!(card->flags & CARD_USEKANJI) && jfc_config.cfg.kanji_unknown)) jfc_window.windows[0].view(true);
           }
         break;
  }
  jfc_window.set (1,false);                     // Set the first window as active.
  if (jfc_config.cfg.time_cards && jfc_config.cfg.time_show) SetTimer (main_window,TIME_SHOW,jfc_config.cfg.time_show,NULL);
  return;
}

//
//  Read a card file.  This is the big routine, this reads the file and intializes the 
//  cards.
//
//      name   -- Name of file to read.
//
//      RETURN -- A non-zero return value indicates an error.
//
int JFC_file::read (TCHAR *name) {
  JFC_card *card;
  JFC_flash *flsh,*f,*ff;
  long i,j,k;
//
//  Setup the file name.
//
  if (!(filename = strdup(name))) { OutOfMemory(); return (true); }
  window_title ();
//
//  Actually read the file as a sub-file.
//
  if (read_subfile(name)) return (true);
//
//  Link subfile cards into one list.
//
  class SUB_file *file;
  for (file = files; file; file = file->next) {
    if (!file->cards) continue;
    if (!cards) cards = file->cards; else card->next = file->cards;
    card = file->last;
  }
//
//  If there is a saved file for this flash set we want to read it, then 
//  jump to setup stats setup
//
  if (read_save()) goto setup_stats;
//
//  Process the card data into actual cards.
//
  for (card = cards; card; card = card->next) {
    if      (card->flags & CARD_SUBCARD  );
    else if (card->flags & CARD_KANJI    ) {
      if (card->flags & (CARD_KANJIONLY | CARD_KANAONLY | CARD_ASCIIONLY)) {
        if (card->flags & CARD_KANJIONLY) flash_add (card,FLASH_KANJIKANJI);
        if (card->flags & CARD_KANAONLY ) flash_add (card,FLASH_KANJIKANA );
        if (card->flags & CARD_ASCIIONLY) flash_add (card,FLASH_KANJIASCII);
      }
      else {
        if (jfc_config.cfg.kanji_ascii) flash_add (card,FLASH_KANJIASCII);
        if (jfc_config.cfg.kanji_kana ) flash_add (card,FLASH_KANJIKANA );
        if (jfc_config.cfg.kanji_kanji) flash_add (card,FLASH_KANJIKANJI);
      }
    }
    else if (card->flags & (CARD_KANJIONLY | CARD_KANAONLY | CARD_ASCIIONLY)) {
      if (card->flags & CARD_KANAONLY ) flash_add (card,FLASH_KANA);
      if (card->flags & CARD_KANJIONLY) flash_add (card,FLASH_KANJI);
      if (card->flags & CARD_ASCIIONLY) flash_add (card,FLASH_ASCII);
    }
    else {
      if (jfc_config.cfg.flash_ascii) flash_add (card,FLASH_ASCII);
      if (jfc_config.cfg.flash_kana ) flash_add (card,FLASH_KANA);
      if (jfc_config.cfg.flash_kanji && (card->flags & CARD_USEKANJI)) flash_add (card,FLASH_KANJI);
    }
  }  
//
//  If the user has used fixed number of cards then we may need to reduce the number of cards.
//
  if (jfc_config.cfg.use_max && (flash_count > jfc_config.cfg.maximum)) {
    flsh        = flash;
    flash       = NULL;
    flash_count = 0;
    for (i = 0; flsh && (i < jfc_config.cfg.maximum); i++) {    // Loop for number cards wanted
      for (j = 0, f = flsh; f; f = f->next) j += f->prob;       // Count weights
      k = random(j);                                            // Pick random number.
      if (k < flsh->prob) {                                     // Picked first card
        f       = flsh;
        flsh    = f->next;
        f->next = flash;
        flash   = f;
        flash_count++;
      }  
      else {                                                    // Picked any other card.
        k -= flsh->prob;
        for (f = flsh; f->next->next && (f->next->prob > k); f = f->next) k -= f->next->prob;
        ff       = f->next;
        f->next  = ff->next;
        ff->next = flash;
        flash    = ff;
        flash_count++;
      }
    }
    while (flsh) {                                              // Delete any unused cards
      f    = flsh;
      flsh = f->next;
      delete f;
    }
  }
//
//  Now randomize the cards.  If the user used a max count this will actually 
//  double randomize, but this is fast.
//
  flsh  = flash;
  flash = 0;
  for (i = flash_count; flsh; i--) {
    j = random(i);
    if (j == 0) {               // First card.
      f       = flsh;
      flsh    = flsh->next;
      f->next = flash;
      flash   = f;
    }
    else {                      // Any other card.
      for (f = flsh; j > 1; j--) f = f->next;
      ff       = f->next;
      f->next  = ff->next;
      ff->next = flash;
      flash    = ff;
    }
  }
//
//  Setup stats
//
  right = wrong = 0;                    // These will be setup by the file reader.
setup_stats:;
//
//  Handle the case where the deck generated no cards
//
  if (!flash_count) {
    static class JFC_card error_card;
    error_card.data = (byte *) ". This Flash Card File Generated no Cards!";
    flash_add  (&error_card,FLASH_ERROR);
    flash_next ();
    JMessageBox (main_window,IDS_FILE_NOCARDS,IDS_FILE_NOCARDSTITLE,MB_OK | MB_ICONWARNING);
  }
//
//  Select the first card.
//
  flash_next ();
  return     (false);
}

//
//  Read saved set of cards.  This is used to recover a set of cards that was in 
//  progress and was saved.
//
//      RETURN -- A non-zero value indicates saved cards were recovered.  This will
//                suppress the card selection and randomization process normally used.
//
//  File format is a header block that contains mostly ID information and the number of
//  cards stored in the file.
//
//  The rest of the file consists of a record per card and are stored in order.  The link
//  in the card data is converted to a zero based number indicating the offset from the 
//  first card data line.
//
//  In reconstruction, the card data is frist read in.  This will place all the card
//  data in the same order it was before (we check the size, and date of the file to 
//  make sure it has not changed).  We can then convert the card data from index offsets
//  into the data back into actual pointers.
//
int JFC_file::read_save () {
  struct FileSaveHeader filehead;   // File header
  struct FileSubFile    filesub;    // Subfile information
  struct FileFlash      fileflash;  // Record for each fflash card
  TCHAR  name[SIZE_BUFFER];         // Name of save file.
  HANDLE file;                      // File handle
  long   i;                         // Scratch
  ulong  s;                         // Used to check for write errors.
  JFC_flash *flsh,*f;               // Manipulates cards list
  JFC_card  *card;                  // Used to find card file data for card.
  SUB_file  *subfile;               // Subfile.
//
//  Open the file and read the contents into a memory buffer.
//
  if (INVALID_HANDLE_VALUE == (file = OPENREAD(save_name(name)))) return (false);
  ReadFile (file,&filehead,sizeof(filehead),&s,NULL);
  if ((filehead.magic != MAGIC_SAVE) || (s != sizeof(filehead))) {
    ErrorMessage (true,IDS_FILE_SAVECORRUP,name);
    CloseHandle  (file);
    return       (false);
  }
//
//  Make sure the card data file has not changed.
//
  for (i = 0, subfile = files; (i < filehead.files) && (subfile); i++, subfile = subfile->next) {
    ReadFile (file,&filesub,sizeof(filesub),&s,NULL);
    if (s != sizeof(filesub)) {
      ErrorMessage (true,IDS_FILE_READERROR,name);
      CloseHandle  (file);
      DeleteFile   (name);
      return       (true);
    }
    if (CompareFileTime(&subfile->time,&filesub.time) || (subfile->size != filesub.size)) {
      if (IDYES != JMessageBox (main_window,IDS_SAVE_TEXT,IDS_SAVE_TITLE,MB_YESNO | MB_ICONWARNING | MB_DEFBUTTON2,subfile->filename)) { 
        CloseHandle (file); 
        DeleteFile  (name);
        return      (true); 
      }
    }
  }
//
//  Process the cards
//
  flash_count = filehead.count;
  right       = filehead.right;
  wrong       = filehead.wrong;
  for (i = 0; i < flash_count; i++) {
    ReadFile (file,&fileflash,sizeof(fileflash),&s,NULL);       // Read
    if (s != sizeof(fileflash)) {
      ErrorMessage (true,IDS_FILE_READERROR,name);
      CloseHandle  (file);
      DeleteFile   (name);
      return       (true);
    }
    if (!(flsh = new JFC_flash)) {                              // Allocate
      OutOfMemory ();
      CloseHandle (file);
      return      (true);
    }                                                           // Copy data and link
    for (card = cards; fileflash.card; fileflash.card--) card = card->next;
    flsh->card      = card;
    flsh->duplicate = fileflash.duplicate;
    flsh->prob      = fileflash.prob;
    flsh->type      = fileflash.type;
    flsh->next      = flash;
    flash           = flsh;
  }
//
//  Reverse the order of the cards.  We read them from the file, but this generates them
//  in reverse order.
//
  flsh    = flash;
  flash   = NULL;
  while (flsh) {
    f       = flsh;
    flsh    = f->next;
    f->next = flash;
    flash   = f;
  }
//
//  Cleanup and remove the save file.  If the user closes again, another save file
//  should be generated.
//
  CloseHandle (file); 
  DeleteFile  (name);
  return      (true);
}

//
//  Read a subfile into the current set of cards.
//
//      name   -- Name of subfile to read.
//
//      RETURN -- A non-zero value indicates and error.
//
int JFC_file::read_subfile (TCHAR *name) {
  int       err;
  SUB_file *file;
  if (count > jfc_config.cfg.max_subfiles) {
    ErrorMessage (true,IDS_FILE_TOOMANY,jfc_config.cfg.max_subfiles);
    return (true);
  }
  if (!(file = new SUB_file)) { OutOfMemory(); return (true); }
  if (last) last->next = file; else files = file;
  last = file;
  count++;
  if ((err = file->load(name,files == file)) && (file != files)) ErrorMessage (false,IDS_FILE_SUBFILELOAD,name);
  return (err);
}

//
//  This rotuine is called as a responce to changing the options.  This gives the user 
//  the option of re-opening the current card set if they want to.
//
//      check -- If set to true, this will prompt the user before reopeingin the file.
//
void JFC_file::reinit (int check) {
  TCHAR buffer[SIZE_BUFFER];
  if (this && flash) {
    if (!check || YesNo (IDS_FILE_SETTINGS)) {
      lstrcpy (buffer,filename);
      close     (false);
      open_file (buffer);
    }
    else {
      flash_next ();
    }
  }
  return;
}

//
//  This routine saves the current set of cards.  This is used to continue a set of 
//  cards from where you left off.
//
//      RETURN -- Non-zero value indicates an error saving.
//
//  File format is a header block that contains mostly ID information and the number of
//  cards stored in the file.
//
//  The rest of the file consists of a record per card and are stored in order.  The link
//  in the card data is converted to a zero based number indicating the offset from the 
//  first card data line.
//
int JFC_file::save () {
  struct FileSaveHeader filehead;   // Header for file
  struct FileSubFile    filesub;    // Header for each subfile.
  struct FileFlash      filecard;   // Record for each card
  class  SUB_file      *subfile;    // Pointer to current subfile
  TCHAR  name[SIZE_BUFFER];         // Name of file.
  HANDLE file;                      // File handle
  ulong  s;                         // Size used to check for errors.
  int    i;                         // Scratch
  int    err = false;               // Write error flag.  Non-zero value inidcates error
  JFC_flash *flsh;                  // For scanning fash cards.
  JFC_card  *card;                  // For scanning cards
//
//  Open the file.
//
  if (INVALID_HANDLE_VALUE == (file = OPENWRITE(save_name(name)))) {
    ErrorMessage (true,IDS_FILE_SAVEOPEN,name);
    return       (true);
  }
//
//  Write header.
//
  filehead.magic = MAGIC_SAVE;          // This is a statistic file.
  filehead.count = flash_count;         // Number of flash cards left to go.
  filehead.right = right;               // Number of correct cards.
  filehead.wrong = wrong;               // Number of wrong cards.
  filehead.files = count;               // Number of subfiles.
  WriteFile (file,&filehead,sizeof(filehead),&s,NULL);
  if (s != sizeof(filehead)) err = true;
//
//  Write information on each subfile.
//
  for (subfile = files; subfile; subfile = subfile->next) {
    filesub.size = subfile->size;
    filesub.time = subfile->time;
    WriteFile (file,&filesub,sizeof(filesub),&s,NULL);
    if (s != sizeof(filesub)) err = true;
  }
//
//  Write each card.
//
  for (flsh = flash; flsh; flsh = flsh->next) {
    for (card = cards, i = 0; flsh->card != card; card = card->next) i++;
    filecard.card      = i;
    filecard.duplicate = flsh->duplicate;
    filecard.type      = flsh->type;
    filecard.prob      = flsh->prob;
    WriteFile (file,&filecard,sizeof(filecard),&s,NULL);
    if (s != sizeof(filecard)) err = true;
  }
//
//  Cleanup and close file.
//
  CloseHandle (file);
  if (err) {
    ErrorMessage (true,IDS_FILE_SAVEWRITE,name);
    return       (true);
  }
  return (false);
}

//
//  This routine builds the save-file name for the current file.  This is basically the main 
//  file name with the extension .jsv.
//
//      buffer -- Buffer to build the file name in (assumed to be of adequate size).
//      
//      RETURN -- Pointer to buffer (just for convenience).
//
TCHAR *JFC_file::save_name (TCHAR *buffer) {
  int i;
  lstrcpy (buffer,filename);
  for (i = lstrlen(buffer)-1; (i >= 0) && (buffer[i] != ':') && (buffer[i] != '//') && (buffer[i] != '.'); i--);
  if (buffer[i] == '.') buffer[i] = 0;
  lstrcat (buffer,TEXT(".jsv"));
  return (buffer);
}

//
//  This routine builds the stat-file name for the current file.  This is basically the main 
//  file name with the extension .jst.  At this point, this routine is only used for the 
//  delete option to delete a file and it's stat-file.
//
//      buffer -- Buffer to build the file name in (assumed to be of adequate size).
//      
//      RETURN -- Pointer to buffer (just for convenience).
//
TCHAR *JFC_file::stat_name (TCHAR *buffer) {
  int i;
  lstrcpy (buffer,filename);
  for (i = lstrlen(buffer)-1; (i >= 0) && (buffer[i] != ':') && (buffer[i] != '//') && (buffer[i] != '.'); i--);
  if (buffer[i] == '.') buffer[i] = 0;
  lstrcat (buffer,TEXT(".jst"));
  return (buffer);
}

//                                                                   //
//  End class JFC_file.
//
//-------------------------------------------------------------------//

//-------------------------------------------------------------------//
//                                                                   //
//  Exported routines.
//

//
//  This routine opens a file with a specific name.
//
//      name   -- Name of file.  If the name field is set to NULL, then
//                a file requestor will be generated.
//
//      RETURN -- A non-zero value indicates an error.  Error message will
//                be displayed by the routine (or called routine).
//
int open_file (TCHAR *name) {
  TCHAR buffer[SIZE_BUFFER];
//
//  Generate a file requestor if necessary.
//
  if (!name) {
    OPENFILENAME ofn;
    TCHAR filters[SIZE_BUFFER];
    memset (&ofn,0,sizeof(ofn));
    ofn.lStructSize       = sizeof(ofn);
    ofn.hwndOwner         = main_window;
    ofn.hInstance         = instance;
    ofn.lpstrFilter       = zero_string(filters,IDS_FILTERS_JFC);
    ofn.nFilterIndex      = 1;
    ofn.lpstrFile         = buffer;
    ofn.nMaxFile          = SIZE_BUFFER;
    ofn.Flags             = OFN_CREATEPROMPT | OFN_PATHMUSTEXIST | OFN_HIDEREADONLY  | OFN_EXPLORER;
#ifdef WINCE
    ofn.lpstrInitialDir   = currentdir;         // Use Current directory for windows CE              
#endif WINCE
    buffer[0] = 0;
    if (!GetOpenFileName(&ofn)) return (false); // User canclled!
#ifdef WINCE
    set_currentdir (buffer,true);               // Set windows CE current directory.
#endif WINCE
    name = buffer;
  }
//
//  Generate a new file structure from which to work.
//
  if (!(jfc_file = new JFC_file)) { OutOfMemory (); return (true); }
//
//  This is a trick to get the this pointer to point to the new item.
//
  if (jfc_file->read(name)) {
    ErrorMessage (true,IDS_FILE_OPEN,name);
    return (true);
  }
  recent_files   (name);
  EnableMenuItem (hmenu,IDM_FILE_CLOSE ,MF_BYCOMMAND | MF_ENABLED);
  EnableMenuItem (hmenu,IDM_FILE_REOPEN,MF_BYCOMMAND | MF_ENABLED);
  return (false);
}

//
//  This routine processes files into the recent files list.  This 
//  routine is called whenever a named file is open or closed.
//
//      name -- Name of file to add to the menu.
//
#define MENU_FILE       0   // Location of the file menu.
#define MENU_RECENT     5   // Location of the recent files sub-menu.

void recent_files (TCHAR *name) {
  int   i,j;
  TCHAR buffer[SIZE_BUFFER];
  HMENU menu;
  menu = GetSubMenu(GetSubMenu(hmenu,MENU_FILE),MENU_RECENT);       // Get recent files sub-menu
  wsprintf (buffer,TEXT("&1 %s"),name);                             // Make menu item
  InsertMenu (menu,0,MF_BYPOSITION,IDM_FILE_FILES_BASE,buffer);     // Insert at the top of the list.
  for (i = 1; j = get_menudata(menu,i,true,buffer); i++) {          // Check the rest of the list.
    RemoveMenu (menu,i,MF_BYPOSITION);                              // Dump item.
    if ((j == IDM_FILE_FILES_NONE) || (i >= 9) || !stricmp(buffer+3,name)) { i--; continue; }
    buffer[1] = '1'+i;                                              // If here we will re-add item.
    InsertMenu (menu,i,MF_BYPOSITION,IDM_FILE_FILES_BASE+i,buffer);
  }
  return;
}

//
//  Puts the title in the window to match the current file.
//
void window_title () {
  int   i;
  TCHAR buffer[SIZE_BUFFER];
#ifdef WINCE_PPC
  if (jfc_file && jfc_file->filename) {
    for (i = lstrlen(jfc_file->filename)-1; i && (jfc_file->filename[i] != '\\'); i--);
    lstrcpy (buffer,jfc_file->filename+i+1);
  }
  else {
    lstrcpy (buffer,VERSION_NAME);
  }
  SetWindowText   (main_window,buffer);
  jfc_stat.redraw ();
#else  WINCE_PPC
  lstrcpy (buffer,VERSION_NAME);
  lstrcat (buffer,TEXT(" -- "));
  if (jfc_file && jfc_file->filename) {
    if (!jfc_config.cfg.short_name) lstrcat (buffer,jfc_file->filename);
      else {
        for (i = lstrlen(jfc_file->filename)-1; i && (jfc_file->filename[i] != '\\'); i--);
        lstrcat (buffer,jfc_file->filename+i+1);
      }
  }
  else {
    lstrcat (buffer,TEXT("(C) glenn rosenthal 1999,2000"));
  }
  SetWindowText   (main_window,buffer);
  jfc_stat.redraw ();
#endif WINCE_PPC
  return;
}






SUB_file::SUB_file () {
  memset (this,0,sizeof(SUB_file));
  return;
}

SUB_file::~SUB_file () {
  class JFC_card  *card;
  if (filename) free (filename);    // Delete file name.
  while (cards) {                   // Delete master cards set.
    card  = cards;
    cards = card->next;
    delete card;
  }
  memset (this,0,sizeof(SUB_file));
  return;
}








