//-------------------------------------------------------------------//
//                                                                   //
//  JFC Copyright (C) Glenn Rosenthal, 1999,2000,2001.               //
//  All rights reserved.                                             //
//                                                                   //
//-------------------------------------------------------------------//

//-------------------------------------------------------------------
//
//  This modlues main function is to deal with the system fonts.  
//  There are three fonts used in the system currently.  These are
//  the kana/kanji font.  The user font used in the main edit window,
//  and the system font used in controls, and in the status bar.  
//  Various useful parameters are caclated buy the font system for use
//  by other routines.  The system currenly allows for changes in the 
//  fonts, but these are relativly limited.  
//
//  The onther function supported in this modlue is the color-kanji
//  feature.  This feature allows the user to generate a list of kanji.
//  When kanji are rendred, the kanji on this list (or off this list)
//  can be rendered in a different color.
//
//  Font system:
//
//  The original JFC fonts come in a number of sizes.  The fonts 
//  soted in the files are filled to a byte, thus a font image that is
//  16x16 requries 32 bytes, and a font that is 24x24 requires 72 bytes.
//  Windows, however, requries that bitmap widths be a multable of 
//  16 bits.  This has major implications when using fonts such as 
//  the 24x24 font.  This font has to be converted to a memory image
//  that is 32 (width)x24 (heigth).  JFC only rescales fonts that 
//  are cached, thus such a font will always be cached.
//
//  Caches:
//
//  JFC allows caching of some fonts.  A font that is cahced, uses
//  the bitmaps field points to the cache.  The cache parameter points
//  to an array of index values (not JIS values, but rather internal 
//  font table values).  These are the cached values.  The index field
//  points to the next location to be replaced when a chache miss occres.
//  The index value is cyclic and simply goes around and around.  This
//  means that even kana will get droped from the cache, but this 
//  seems to work well.  JFC keeps the font file open if the file
//  is cached.  Thus we don't have to reopen it.  (As usuall, the 
//  file is open is share mode.)
//  

#include "jfc_main.h"
#include "jfc_conf.h"
#include "jfc_font.h"
#include "jfc_jisc.h"
#include "jfc_misc.h"
#include <winnls.h>

//-------------------------------------------------------------------
//
//  Compile-time options.
//

#define NAME_KANJI      TEXT("k16x16.f00")  // Name of default system font.
#define NAME_BIGFONT    TEXT("k48x48.f00")  // Name of big kanji font.
#define NAME_MEDFONT    TEXT("k24x24.f00")  // Name of medium kanji font.

#define SIZE_KANJIBUFFER    1026            // Implies a 128x128 max font size.

//-------------------------------------------------------------------
//
//  Static routines.
//
//  Provide various reveice capabilities for routines in this module.
//

//-------------------------------------------------------------------
//
//  Data and definitions assoicated with color-kanji.
//
#define COLORKANJI_NAME TEXT("colkanji.lst")    // File name for storeing color-kanji list.

#define TEXT_COLORKANJI "Color-Kanji List"  // Text for error messages, just to make sure is it the same.

static KANJI *colorkanji_list = NULL;              // Pointer to the actual list.
static short  colorkanji_size = 0;                 // Size of the list in kanji.

//-------------------------------------------------------------------
//
//  Static routines.
//

//
//  Determines if a kanji is in a specific set.
//
//      kanji  -- Kanji to search for.
//      set    -- Pointer to list of kanji that is the list.
//      size   -- Number of characterrs in the list.
//
//      RETURN -- A return value of zero indicates that the kanji is
//                not in the list.  Any other return is 1+the index 
//                of the character in the list.
//
static int in_set (int kanji,KANJI *set,int size) {
  int i;
  for (i = 0; i < size; i++) if (kanji == set[i]) return (i+1);
  return (false);
}

//
//  This is a faster version of open_font() that is optimized for opening a bitmapped font.
//  This was separated out to save some code space.
//
//      name   -- Name of font to open.
//      cache  -- Indicates if the font is to be cahced.
//
//      RETURN -- Pointer to newly open bitmapped font or NULL, if the font could not be open.
//
static KANJI_font *open_bitmap (TCHAR *name,int cache) {
  BITMAP_KANJI_font *font;
  if (!(font = new BITMAP_KANJI_font)) return (NULL);
  if (font->open(name,cache)) {
    font->remove ();
    font = NULL;
  }
  return (font);
}

//
//  Attempts to open a font based on a name, and size.  This routine autmatically handles
//  both bitmaped and TrueType fonts.
//
//      name     -- Name of font to open.  If this ends in .f00 it is assumed to be a 
//                  bitmapped font.  Any other ending is a TrueType font.
//      size     -- Size of the font in pixels (used only for TrueType fonts).
//      cache    -- Indictes if the font is to be cached (used only for bitmapped fonts).
//      hdc      -- If included indicates the DC for which the font will be used.  If this is
//                  not included, we autmatically assume it is a screen font.
//
//      RETURN -- Pointer to an open font, or NULL if an error occures.
//
static KANJI_font *open_font (TCHAR *name,int size,int cache,HDC hdc) {
  KANJI_font *font;
  if ((lstrlen(name) >= 4) && stricmp(name+lstrlen(name)-4,TEXT(".f00"))) {
    HDC display_hdc = NULL;
    TRUETYPE_KANJI_font *tt_font;
    if (!hdc) display_hdc = hdc = GetDC(main_window);       // No DC so this is a screen font, make a DC.
    if (!(tt_font = new TRUETYPE_KANJI_font)) return (NULL);// Allocate class
    if (tt_font->open(hdc,name,size)) {                     // Try to open
      tt_font->remove();
      tt_font = NULL;
    }
    font = tt_font;
    if (display_hdc) ReleaseDC (main_window,hdc);           // If we made the dc then get rid of it
  }
  else {
    font = open_bitmap(name,cache);
  }
  return (font);
}

//-------------------------------------------------------------------
//
//  Begin Class KANJI_font.
//
//  This class contains most of the kanji forn manimulation and display 
//  code.
//

#define NO_CHANGE   0xFFFFFFFF  // Indicates no change for the color (This color is invalid in Windows).

static class KANJI_font *big_kanji   = NULL;    // Largest font avaialbe, used for big font in character info.

//
//  This routine sets the display color based on if a character is in the color-kanji list.
//
//      jis   -- Character to be checked (JIS code).
//      hdc   -- Display context (since this may change the text color).
//      color -- If the color is unchanged, this will return as -1.  If the color is changed,
//               this will return as the original color so that it may be restored.
//
void KANJI_font::find_color (int jis,HDC hdc,COLORREF &color) {
  int i;
  color = -1;
  if (jfc_config.cfg.colorkanji_mode && colorkanji_list && ISKANJI(jis)) {
    i = in_set(jis,colorkanji_list,colorkanji_size);
    if (jfc_config.cfg.colorkanji_mode == COLORKANJI_MATCH) {
      if ( i) color = SetTextColor(hdc,jfc_config.cfg.colorkanji_color);
    }
    else {
      if (!i) color = SetTextColor(hdc,jfc_config.cfg.colorkanji_color);
    }
  }
  return;
}

//
//  This is a general routine to close and deallocate any resource associated with a font.
//  This will work on any supported font type, and will even handle the case of the font
//  already being closed.
//
void KANJI_font::remove () {
  if (!this) return;            // Font is not open
  close ();                     // Close font
  delete this;                  // Delete font class item.
  return;
}

//-------------------------------------------------------------------
//
//  Begin Class BITMAP_KANJI_font.
//
//  This class contains the derived class for dealing with bitmapped kanji fonts.  These are
//  the fonts distributed with JFC/JFC as apposed to TrueType fonts.
//

typedef struct {                // Header structure from font file.
    KANJI facename[20];         // Font family name
    INT16 width, height;        // Width and height in pixels 
    INT16 charsize;             // Number of bytes that make up a character 
    INT16 verticals;            // The number of extra "vertical" characters
    long  offset;               // Byte offset to the beginning of bitmaps 
    INT16 holes;                // Packed JIS coding or 93x93 with holes
    INT16 leading, spacing;     // Horizontal / Vertical gaps
    char extra[6];              // Make it up to 64 bytes  
} FONTHEADER;

//
//  Simple constructor that simply clears some of the parameters.
//
BITMAP_KANJI_font::BITMAP_KANJI_font () {
  bitmaps = NULL; 
  hdcmem  = NULL;
  cache   = NULL;
  file    = INVALID_HANDLE_VALUE;
#ifndef WINCE
  hbitmap = NULL; 
#endif
  return;
}

//
//  Close a font.
//
void BITMAP_KANJI_font::close () {
  if (bitmaps) free         (bitmaps);
  if (hdcmem ) DeleteDC     (hdcmem);
  if (cache  ) free         (cache);
  if (file != INVALID_HANDLE_VALUE) CloseHandle (file);
  bitmaps  = NULL;
  hdcmem   = NULL;
  cache    = NULL;
  file     = INVALID_HANDLE_VALUE;
#ifndef WINCE
  if (hbitmap) DeleteObject (hbitmap);
  hbitmap = NULL;
#endif
  return;
}

//
//  The big one here, this is the kanji rendering engine:
//
//      hdc -- Context to render into.
//      jis -- JIS code to render (will be converted on the fly).
//      x,y -- Location in the context.
//     
void BITMAP_KANJI_font::draw (HDC hdc,int jis,int x,int y) {
  long     index;
  COLORREF color;
  index  = find_kanji(jis,hdc,color);
#ifdef WINCE
  HBITMAP hbitmap,tbitmap;
  hbitmap = CreateBitmap(width,height,1,1,bitmaps+index);
  tbitmap = SelectObject (hdcmem,hbitmap);
  BitBlt       (hdc,x+hshift,y-height,width,height,hdcmem,0,0,SRCAND);
  SelectObject (hdcmem,tbitmap);
  DeleteObject (hbitmap);
#else
  SetBitmapBits (hbitmap,bmsize,bitmaps+index);
  BitBlt        (hdc,x+hshift,y-height,width,height,hdcmem,0,0,SRCAND);
#endif WINCE
  if (color != -1) SetTextColor(hdc,color);
  return;
}

//
//  This is a variant on the KANJI_font::draw routine that streatches
//  the font to fill a given rectange.  This is used for the big 
//  character in the character-info dialog box.
//
//      hdc      -- Context to render into.
//      jis      -- JIS code to render (will be converted on the fly).
//      rect     -- Rectangle that the character should fill to.
//
void BITMAP_KANJI_font::fill (HDC hdc,int jis,RECT *rect) {
  char    *bitmap;                      // The character bitmap
  int      x,y;                         // Character shifts
  long     index;                       // Index (byte) into bitmap arrays
  COLORREF color;                       // Save color for color kanji.
  index  = find_kanji(jis,hdc,color);   // Find the chracter.
  x = y = 0;                            // Character shift is normally zero.
  bitmap = bitmaps+index;               // Get the character bitmap
//
//  Back to general rotuine.
//
//  SetStretchBltMode (hdc,HALFTONE);   //### These do not appear to change the apperance any
//  SetBrushOrgEx     (hdc,0,0,NULL);
#ifdef WINCE
  HBITMAP hbitmap,tbitmap;
  hbitmap = CreateBitmap(width,height,1,1,bitmap);
  tbitmap = SelectObject(hdcmem,hbitmap);
  StretchBlt   (hdc,rect->left-x,rect->top-y,rect->right-rect->left,rect->bottom-rect->top,hdcmem,0,0,width,height,SRCAND);
  SelectObject (hdcmem,tbitmap);
  DeleteObject (hbitmap);
#else  WINCE
  SetBitmapBits (hbitmap,bmsize,bitmap);
  StretchBlt    (hdc,rect->left-x,rect->top-y,rect->right-rect->left,rect->bottom-rect->top,hdcmem,0,0,width,height,SRCAND);
#endif WINCE
  if (color != NO_CHANGE) SetTextColor(hdc,color);
  return;
}

//
//  This routine finds a kanji value in the font cache.  If the value 
//  is not already loaded, this rotuine loads it.
//
//      jis    -- Character being requirested.
//      hdc    -- Device context to redner character into
//      color  -- Color setting register, if set to -1 thi is not 
//                using color kanji.  If this is set to any other 
//                color this is the saved version of color fonts.
//
//      RETURN -- Index into font bitmaps for this character.  This 
//                is not the character index, but rather the index 
//                to the actual bitmap.
//
int BITMAP_KANJI_font::find_kanji (int jis,HDC hdc,COLORREF &color) {
  int i;
  unsigned long done;
//
//  Is this character in the color-kanji list.
//
  find_color (jis,hdc,color);
//
//  Now find the character.
//
  jis = jis_index(jis);                     // Convert to intial font code.
  if (!cache) return (jis*bmsize);          // Font is not chached.
//
//  Search cache.
//
  if (i = in_set(jis+1,(KANJI *) cache,jfc_config.cfg.font_cache)) return (--i*bmsize);
//
//  Not in cache, so we need to load the character.
//
  SetFilePointer (file,sizeof(FONTHEADER)+jis*bmfilesize,NULL,FILE_BEGIN);
//
//  Word aligned font, needs no translation.
//
  if (bmsize == bmfilesize) ReadFile (file,bitmaps+index*bmsize,bmsize,&done,NULL);
//
//  Non-word aligned font (implies byte alighed), needs translation.
//
    else {
      int   j;
      char *ptr,*p,buffer[SIZE_KANJIBUFFER];
      ptr = bitmaps+index*bmsize;                       // Memory font data loation.
      ReadFile (file,buffer,bmfilesize,&done,NULL);     // Read file font data.
      for (p = buffer, i = 0; i < height; i++) {        // Do rows.
        for (j = 0; j < width/8; j++) *ptr++ = *p++;    //   Do columns
        *ptr++ = 0;                                     //   Extra byte per line.
      }
    }
//
//  Update cache and return.
//
  cache[index] = jis+1;
  jis          = index++;
  if (index == jfc_config.cfg.font_cache) index = 0;
  return (jis*bmsize);
}

//
//  This routine converts between JIS codes used in the 
//  program and internal codes used in the fonts.  Note
//  the font information is highly compresses, so unused 
//  values have been removed.
//
//  I got this routine from JFC's soruce, and the following is the 
//  notice contained in the JFC source file:
//
//  Copyright (C) Stephen Chung, 1995.  All rights reserved. 
//  The following routine came from KD, the Kanji Driver program  
//  written by Izumi Ozawa.  It was reformatted and slightly      
//  modified.  Many thanks to Izumi for his donation.            
//
//  Note, I have modified the code yet again, but only slightly.
//
int BITMAP_KANJI_font::jis_index (int jis) {
  int hi, lo;

  hi = HIBYTE(jis);   // Removed & 0x007f, because, should be unecessary.
  lo = LOBYTE(jis);   // Removed & 0x007f, because, should be unecessary.

  if ((lo <= 0x20) || (lo >= 0x7f) || (hi <= 0x20) || (hi >= 0x75)) goto BadKanji;
  if ((hi == 0x74) && (lo > 0x24)) goto BadKanji;

  hi = 94 * (hi - 33) + (lo - 33);

  if (holes) return (hi);

  /* Skip the holes */

  if      (hi <= 107)                return (hi);      /* 0..93               $2121-$217e, 0  -93   */
                                                       /* 94..107             $2221-$222e, 94 -107  */
  else if (hi >= 203  && hi <=  212) return (hi-56);   /* 203..212   0 .. 9,  $2330-$2339, 147-156  */
  else if (hi >= 220  && hi <=  245) return (hi-63);   /* 220..245   A .. Z,  $2341-$235a, 157-182  */
  else if (hi >= 252  && hi <=  277) return (hi-69);   /* 252..277   a .. z,  $2361-$237a, 183-208  */
  else if (hi >= 282  && hi <=  364) return (hi-73);   /* 282..364  hiragana, $2421-$2473, 209-291  */
  else if (hi >= 376  && hi <=  461) return (hi-84);   /* 376..461  katakana, $2521-$2576, 292-377  */
  else if (hi >= 470  && hi <=  493) return (hi-92);   /* 470..493   GREEK,   $2621-$2638, 378-401  */
  else if (hi >= 502  && hi <=  525) return (hi-100);  /* 502..525   greek,   $2641-$2658, 402-425  */
  else if (hi >= 564  && hi <=  596) return (hi-138);  /* 564..596   RUSSIAN, $2721-$2741, 426-458  */
  else if (hi >= 612  && hi <=  644) return (hi-153);  /* 612..644   russian, $2751-$2771, 459-491  */
                                                       /*  n/a       linedraw,  n/a      , 492-523  */
  else if (hi >= 1410 && hi <= 4374) return (hi-886);  /* 1410..4374 kanji-1, $3021-$4f53, 524-3488 */
  else if (hi >= 4418 && hi <= 7805) return (hi-928);  /* 4418..7805 kanji-2, $5021-$7424, 3490-6877*/

    /* No such kanji */
BadKanji:
  return (jis_index(KANJI_BAD));
}

//
//  This routine opens a font up for operations.
//
//      name    -- Name of font to open.
//      docache -- If non-zero forces font chaching.  
//
//      RETURN  -- A non-zero value indicates an error.
//
//  Not that even if no font caching is requested, JFC may overide
//  this request if the font requested has a non-word alighed strucure.
//  Then to perform the translations, JFC will implement caching
//  anyway.
//
int BITMAP_KANJI_font::open (TCHAR *name,int docache) {
  int number;
  FONTHEADER header;
  unsigned long done;
  if (INVALID_HANDLE_VALUE == (file = jfc_config.open(name,OPEN_READ))) return (true);
  if (!ReadFile(file,&header,sizeof(header),&done,NULL)) return (true);
  truetype   = false;
  width      = header.width;
  height     = header.height;
  bmsize     = header.height*sizeof(short)*((header.width/8+1)/sizeof(short));
  bmfilesize = header.charsize;
  holes      = header.holes;
  leading    = header.leading;
  spacing    = header.spacing/2;                // ### this was modified because character spacing was too big
  hshift     = spacing/2;
  number     = (GetFileSize(file,NULL)-sizeof(header))/bmsize;
  if (bmsize != bmfilesize) docache = true;     // Because of font adjustment, odd font sizes must be cached.
//
//  Cached fonts, allocate tables, and intialize values.
//
  if (docache) {
    cache   = (KANJI *) calloc(jfc_config.cfg.font_cache,sizeof(short));
    bitmaps = (char  *) calloc(jfc_config.cfg.font_cache,bmsize);
    if (!cache || !bitmaps) return (true);
    index = 0;
  }
//
//  Non-cached fonts.  Read entire font table, and clear some values.
//
  else {
    if (!(bitmaps = (char *) calloc(number,bmsize)) || !ReadFile(file,bitmaps,number*bmsize,&done,NULL)) return (true);
    CloseHandle (file);
    file = INVALID_HANDLE_VALUE;
  }
//
//  This stuff was orriginally located in the draw, routine.  This allowed
//  draw to respond to different hdc's.  I moved the standard display
//  stuff here to save on operating speed.
//
//  This stuff builds a memory HDC that is compatable with the screen.
//  Bitmaps are first transfered here then blit into the main display.
//
#ifndef WINCE
  if (!(hbitmap = CreateBitmap(width,height,1,1,bitmaps))) return (true);
#endif WINCE
  HDC hdc;
  hdc    = GetDC(main_window);
  hdcmem = CreateCompatibleDC(hdc);
#ifndef WINCE
  SelectObject (hdcmem,hbitmap);
  SetMapMode   (hdcmem,GetMapMode(hdc));
#endif WINCE
  ReleaseDC    (main_window,hdc);
  return (false);
}

//
//  End Class BITMAP_KANJI_font
//
//-------------------------------------------------------------------

//-------------------------------------------------------------------
//
//  Begin Class TRUETYPE_KANJI_font.
//
//  This class contains the derived class for dealing with TrueType kanji fonts.
//
#ifndef WINCE       // Windows CE does not support Japanese TrueType fonts

#define X   ((char *) &x)

//
//  Swap byte order within a long int.
//
//      x      -- Integer.
//
//      RETURN -- The swapped integer
//
static ulong swap_long (ulong x) {
  char p;
  p    = X[0];
  X[0] = X[3];
  X[3] = p;
  p    = X[1];
  X[1] = X[2];
  X[2] = p;
  return (x);
}

//
//  Swap byte order within a short int.
//
//      x      -- Integer.
//
//      RETURN -- The swapped integer
//
static ushort swap_short (ushort x) {
  char p;
  p    = X[0];
  X[0] = X[1];
  X[1] = p;
  return (x);
}

#endif WINCE

//
//  Simple class constructor to make sure the object is initialized correctly.
//
TRUETYPE_KANJI_font::TRUETYPE_KANJI_font (void) {
  font = NULL;
#ifndef WINCE
  cmap = NULL;
  gsub = NULL;
#endif WINCE
  return;
}

//
//  Close a font.
//
void TRUETYPE_KANJI_font::close () {
  if (font) DeleteObject (font);
  font = NULL;
#ifndef WINCE
  if (cmap) free         (cmap);
  if (gsub) free         (gsub);
  cmap = NULL;
  gsub = NULL;
#endif WINCE
  return;
}

//
//  Basic render routine for TrueType fonts.
//
//      hdc -- Display contect to render to.
//      jis -- Character to render
//      x,y -- Location in the context.
//
void TRUETYPE_KANJI_font::draw (HDC hdc,int jis,int x,int y) {
  COLORREF color;
  ushort  glyph;

  find_color (jis,hdc,color);                       // Get color.
  SetBkMode (hdc,TRANSPARENT);                      // Make background transparent
  font = (HFONT) SelectObject(hdc,font);            // Change fonts
#ifndef WINCE
  glyph = jis_index(jis);                           // Find glyph index
  ExtTextOut (hdc,x+hshift,y+vshift,ETO_GLYPH_INDEX,NULL,(TCHAR *) &glyph,1,NULL);
#else WINCE
  glyph = jis2unicode(jis);
  ExtTextOut (hdc,x+hshift,y+vshift,0,NULL,(TCHAR *) &glyph,1,NULL);
#endif WINCE
  font = (HFONT) SelectObject(hdc,font);            // Restore font
  if (color != -1) SetTextColor(hdc,color);         // Restroe color
  return;
}

//
//  This is a variant on the KANJI_font::draw routine that streatches
//  the font to fill a given rectange.  This is used for the big 
//  character in the character-info dialog box.
//
//      hdc      -- Context to render into.
//      jis      -- JIS code to render (will be converted on the fly).
//      rect     -- Rectangle that the character should fill to.
//
void TRUETYPE_KANJI_font::fill (HDC hdc,int jis,RECT *rect) {
  draw (hdc,jis,rect->left+(rect->right-rect->left+1-width)/2-hshift,rect->bottom-(rect->bottom-rect->top+1-height)/2);
  return;
}

//
//  Convert JIS character code into an glyph index for the font.
//
//      jis    -- JIS code to be processed.
//
//      RETURN -- Glyph index to be used in rendering the character.  If the glyph is not
//                present, the routine will automatically return the default blank character.
//
int TRUETYPE_KANJI_font::jis_index (int jis) {
#ifndef WINCE
  int i;
  jis = jis2unicode(jis);                                       // Convert to UNICODE
  for (i = 0; (i < count) && (jis > end[i]); i++);              // Find segment containning charcter
  if (jis < start[i]) return (jis_index(KANJI_BAD));            // Character is out of segment (does not exist).
  if (!offset[i]) jis += delta[i];                              // Displacement character
    else  jis = *(offset[i]/2 + (jis - start[i]) + &offset[i]); // Get value from glyph array.
  if (!jis) return (jis_index(KANJI_BAD));                      // No not a good character.
  if (gsub) {                                                   // If this is here we are printing vertical.
    for (i = 0; (i < vcount) && (from[i] < jis); i++);          // Find the glyph in the substitution list.
    if (from[i] == jis) jis = to[i];                            // Do we need to stubstitue.
  }
#endif WINCE
  return (jis);                                                 // Return value.
}

//
//  This routine opens a font up for operations.
//
//      hdc      -- Display context for the font (screen or printer).
//      name     -- Name of font to open.
//      size     -- Indicates pixal size of the characters generated by the font.  This is 
//                  ignored for bitmap fonts, but actually determines the font size for TrueType
//                  fonts.
//
//      RETURN   -- A non-zero value indicates an error.
//
#define CMAP (*((ulong *) "cmap"))      // Character map tag
#define GSUB (*((ulong *) "GSUB"))      // Glyph substitution tag
#define VERT (*((ulong *) "vert"))      // Vertical printing tag

//
//  Macro used for addressing within the TrueType font structure.  This calucates and address from
//  the base of one structure and casts the result to the final type.  Note the offset can only
//  be short, since the swap is done here.
//
#define ADDRESS(t,b,o)    (struct t *) (((char *) b)+swap_short(o))
                                                                    
int TRUETYPE_KANJI_font::open (HDC hdc,TCHAR *name,int size) {
//
//  Open the kanji font.
//
  LOGFONT    lf;
  memset (&lf,0,sizeof(lf));
  lstrcpy (lf.lfFaceName,name);
  lf.lfHeight     = -size;
  lf.lfCharSet    = SHIFTJIS_CHARSET;
//lf.lfEscapement = lf.lfOrientation = 0(vertical ? 900 : 0);
  if (!(font = CreateFontIndirect(&lf))) return (true);
  truetype = true;
//
//  Access the font thrugh the DC
//
  TEXTMETRIC tm;
  font = (HFONT) SelectObject(hdc,font);
//
//  Get the metric information from the font.
//
  GetTextMetrics (hdc,&tm);
//width   = (short) tm.tmAveCharWidth;
//height  = (short) tm.tmAscent;
//leading = (short) tm.tmExternalLeading;
  width   = (short) tm.tmMaxCharWidth;
  height  = size; 
  leading = height/8; 
  spacing = width/12;
  hshift  = spacing/2-1;            // This -1 seems to give better screen display and should not show up on the printer.
  vshift  = (short) -height;
//
//  Read the cmap inforamtion from the font.  
//
//  The system must have a way to do this.  This should not have to be done here, but I 
//  could not find another way to do this.
//
#ifndef WINCE
  struct cmap_header {      // cmap header structure
    ushort version;         //   file version number (not used).
    ushort number;          //   Number of encodings
  };
  struct table {            // Encoding sub-tables 
    ushort platform_id;     //   platform id (we want 3 == PC)
    ushort encoding_id;     //   encoding id (we want 1 == UNICODE)
    ulong  offset;          //   offset from start of cmap to the actual sub-table
  } *tables;
  struct subtable {         // A partial version of a cmap sub-table
    ushort format;          //   format (should be 4)
    ushort length;          //   length of sub-table
    ushort version;         //   version (not used)
    ushort segCountX2;      //   twice the number of segments in the sub table 
    ushort searchRange;     //   junk
    ushort entrySelector;   //   junk
    ushort rangeShift;      //   junk
    ushort endCount;        //   start of end list table 
  } *sub;                   // -- Rest of table is calculated on the fly, since this table cannot be represented with a C++ structure.
  int  i;
  long j;

  j = GetFontData (hdc,CMAP,0,NULL,0);                              // Size of the cmap data
  if (!(cmap = (char *) malloc(j))) return (true);                  // Make a memory block for this
  if (GDI_ERROR == GetFontData(hdc,CMAP,0,cmap,j)) return (true);   // Read the cmap.
  j      = swap_short(((struct cmap_header *) cmap)->number);       // Get number of sub-tables.
  tables = (struct table *) (cmap+sizeof(struct cmap_header));      // Get sub-tables
  for (i = 0; i < j; i++) {                                         // Find our sub-table (3,1)
    if ((swap_short(tables[i].platform_id) == 3) && (swap_short(tables[i].encoding_id) == 1)) break;
  }
  if (i >= j) return (true);                                        // Could not find sub-table
  sub = (struct subtable *) (cmap+swap_long(tables[i].offset));     // Find the actual sub-table
  if (swap_short(sub->format) != 4) return (true);                  // Check, is the table format 4.
  count  = swap_short(sub->segCountX2)/2;                           // Get count (number of segments)
  end    = &sub->endCount;                                          // Get address of segment tables
  start  = end+count+1;
  delta  = start+count;
  offset = delta+count;
  glyph  = offset+count;                                            // Get address of the glyph array
  j = swap_short(sub->length)/2;                                    // Swap bytes in the sub-table
  for (i = 0; i < j; i++) ((ushort *) sub)[i] = swap_short(((ushort *) sub)[i]);
#endif WINCE
//
//  Clean-up and restore the default font.
//
  font = (HFONT) SelectObject(hdc,font);
  return (false);
}

//
//  End Class TRUETYPE_KANJI_font
//
//-------------------------------------------------------------------

//-------------------------------------------------------------------
//
//  Begin Class JFC_font.
//
//  This class contains information ralted to fonts used by the program,
//  there sises, and other features.  Much of the information contained
//  in this class is simply mainted such that other parts of the 
//  program do not need to recalculate these values.
//

class JFC_font sys_font;        // System font
class JFC_font kanji_font;      // Kanji display font
class JFC_font kana_font;       // Kana display font
class JFC_font imi_font;        // Meaning display font.

//
//  Close the font class and deallocate all resources.
//
void JFC_font::close () {
  if (!duplicate) {
    kanji->remove ();
    if (ascii) DeleteObject (ascii);
  }
  ascii = NULL;
  kanji = NULL;
  return;
}

//
//  Generates a copy of the font.  This is used because the system font and the 
//  display font are speparate.  If the user wants to use the same font for each
//  we just copy the font.  Thus we have only one open copy of the font at a
//  time.  
//
//  This routine just duplicates the JFC_font structure and sets the duplicate flag.
//
//      font -- Font to be copied
//
void JFC_font::copy (class JFC_font *font) {
  *this     = *font;
  duplicate = true;
  return;
}

//
//  Calculate horizontal advance for a character.
//
//      x      -- Inital horizontal position.
//      ch     -- Character.
//
//      RETURN -- New hornizontal poisiton.
//
int JFC_font::hadvance (int x,int ch) {
  if (ISJIS(ch)) return (x+hwidth);
  if (ch != '\t') return (x+widths[ch]);
  return (((x-x_offset)/hwidth+1)*hwidth+x_offset);
}

//
//  This routine opens font system.  This includes a kanji font and an ASCII
//  font of matched size.  All metrics and dimensions are intialized.
//
//      name   -- Name of font.  If the font ends in ".f00" this is assumed
//                to be a bitmapped font.  Otherwise it is assumed to be a 
//                TrueType font.
//      size   -- Height in pixels of the font.
//      cache  -- Non-zero for a chached font.  (Only used for bitmapped fonts.)
//      hdc    -- Display context.  (A value of NULL will assume the main display.)
//
//      RETURN -- Non-zero value indicates success.
//
int JFC_font::open (TCHAR *name,int size,int cache,HDC hdc) {
//
//  Close any open font.
//
  close ();
  duplicate = false;
//
//  Try to open the kanji font.
//
  if (!(kanji = open_font(name,size,cache,hdc))) return (false);
  height   = kanji->height;
  rheight  = kanji->height+1;
  vspace   = height/4;
  hspace   = kanji->width/4;
  hwidth   = kanji->width +  kanji->spacing;
  vheight  = kanji->height+2*kanji->leading;    // This needs to be corrected to get correct line spacing.
  cheight  = vheight-kanji->leading;            // Height of caret (cursor).
  lheight  = vheight;                           // Height of line in a Japanese list box
  loffset  = lheight-height-1;                  // Offset for rendering a line in a list box (for speed this is calcuated here but used only in one place).
  x_offset = hspace;                            // Parameters used JFC_file class for display.
  y_offset = height+vspace;
//
//  Now open the ASCII font
//
  LOGFONT    lf;
  HFONT      tfont;
  TCHAR      string[2];
  SIZE       s;
  int        i;
  memset  (&lf,0,sizeof(lf));
  lstrcpy (lf.lfFaceName,jfc_config.cfg.ascii_font.name);
  lf.lfHeight = -height;
  if (!(ascii = CreateFontIndirect(&lf))) return (false);
//
//  Get font metrics
//  
  hdc = GetDC (main_window);
  tfont = (HFONT) SelectObject (hdc,ascii);
  for (i = 0; i < 256; i++) {                   // Get width occupied by each ASCII
    string[0] = i;                              //   character.
    GetTextExtentPoint32 (hdc,string,1,&s);
    widths[i] = (short) s.cx;
  }
  SelectObject (hdc,tfont);
  ReleaseDC (main_window,hdc);
  return (true);
}

//
//  End Class JFC_font
//
//-------------------------------------------------------------------

//-------------------------------------------------------------------
//
//  Exported data.
//
short sysfont_height;

//-------------------------------------------------------------------
//
//  Exported routines.
//

//
//  This routine deallocates all kanji fonts.  This is called when we are about to exit and 
//  when the user changes the font base, since this will require re-evaluating the font 
//  choics.
//
//      full -- Causes the color kanji list to allso be deallocated.
//
void free_fonts (int full) {
  big_kanji->remove ();
  sys_font  .close  ();
  kanji_font.close  ();
  kana_font .close  ();
  imi_font  .close  ();
  big_kanji = NULL;
  if (full) {
    if (colorkanji_list) free (colorkanji_list);
    colorkanji_list = NULL;
  }
  return;
}

//
//  This routine gets the bigest kanji font avilable.  This is ued 
//  for rendering the large kanji in the characrter info dialog.
//
//      rect   -- Rectangle space.  This is used to determine the font size generated 
//                for TrueType fonts.  This is unused if the main font is not a TrueType 
//                font.
//
//      RETURN -- Pointer to the kanji font.  If necessary, this 
//                rotuine will open a new font, but if it cannot find
//                a font to open, it will return the main system font.
//
KANJI_font *get_bigfont (RECT *rect) {
  int i,j;
  TCHAR *ptr;
//
//  Already have a big font.
//
#ifndef WINCE
  if (big_kanji) return (big_kanji);
#else  WINCE
  if (big_kanji) {
    if (!big_kanji->truetype) return (big_kanji);   // For CE machines, we close and reopen
      else {                                        //   the big font if it is TrueType.  This
        big_kanji->remove();                        //   takes care of problems with the font 
        big_kanji = NULL;                           //   size when opening the kanji view 
      }                                             //   window.
  }
#endif WINCE
//
//  Calcualte font size.
//
  j = rect->right-rect->left;
  i = rect->bottom-rect->top;
  if (j < i) i = j;
//
//  Try to open the font.
//
  if (!jfc_config.cfg.big_font.automatic) {     // User has a perfered big font
    if (big_kanji = open_font(jfc_config.cfg.big_font.name,i,true,NULL)) return (big_kanji);
  }
  if      (!jfc_config.cfg.kanji_font.automatic) ptr = jfc_config.cfg.kanji_font.name;
  else if (!jfc_config.cfg.kana_font .automatic) ptr = jfc_config.cfg.kana_font.name;
  else                                           ptr = jfc_config.cfg.sys_font.name; 
  if (kanji_font.kanji->truetype && rect) {     // TrueType font, so try to make a font to fit in box.
    if ((big_kanji = open_font(ptr,i,true,NULL))) return (big_kanji);
  }
  if (sys_font.kanji->truetype && rect) {       // TrueType font, so try to make a font to fit in box.
    if ((big_kanji = open_font(ptr,i,true,NULL))) return (big_kanji);
  }
  if ((big_kanji = open_bitmap(NAME_BIGFONT,true))) return (big_kanji); // Try 48x48
  if ((big_kanji = open_bitmap(NAME_MEDFONT,true))) return (big_kanji); // Try 24x24
  if (kanji_font.kanji->truetype) {
    if ((big_kanji = open_bitmap(NAME_KANJI,true))) return (big_kanji); // Try 16x16;
  }
  return (kanji_font.kanji);                    // Use the standard font.
}

//
//  Determines if a knaji is in the color-kanji set.
//
//      ch     -- Kanji to be examined.
//
//      RETURN -- Zero indicates the kanji is not in the color kanji set.  Any other 
//                return value is one more than the index into the  color kanji set.
//
int in_colorkanji (int ch) {
  return (in_set(ch,colorkanji_list,colorkanji_size));
}

//
//  This routine initilizes the basic fonts.  It is called during startup and 
//  when the configuration is changed.  
//
//      RETURN -- A non-zero value indicates an error.
//
int initialize_fonts () {
  HDC        hdc;
  TEXTMETRIC tm;
  int        i;
//
//  May be a reinitalize, so we just close some optional fonts.  When
//  they are used again they will be open with the correct parameters.
//
  free_fonts (false);
//
//  Open the display font
//
  if (!sys_font.open(jfc_config.cfg.sys_font.name,jfc_config.cfg.sys_font.size,jfc_config.cfg.cache_systemfont,NULL)) {
    if (!(sys_font.open(NAME_KANJI,16,jfc_config.cfg.cache_systemfont,NULL))) return (true);    // If we cannot open the user's font
  }                                                                                             //   try the default bitmapped font
//
//  Generate the display font.
//
  if (jfc_config.cfg.kana_font.automatic  || !kana_font .open(jfc_config.cfg.kana_font.name ,jfc_config.cfg.kana_font .size,jfc_config.cfg.cache_displayfont,NULL)) kana_font .copy (&sys_font );
  if (jfc_config.cfg.kanji_font.automatic || !kanji_font.open(jfc_config.cfg.kanji_font.name,jfc_config.cfg.kanji_font.size,jfc_config.cfg.cache_displayfont,NULL)) kanji_font.copy (&kana_font);
  if (jfc_config.cfg.imi_font.automatic   || !imi_font  .open(jfc_config.cfg.imi_font.name  ,jfc_config.cfg.imi_font.  size,jfc_config.cfg.cache_displayfont,NULL)) imi_font  .copy (&kana_font);
//
//  Get system font height.
//
  hdc = GetDC (main_window);
  GetTextMetrics (hdc,&tm);                     // Get system font info for later.
  ReleaseDC (main_window,hdc);
  sysfont_height = (short) tm.tmHeight;
//
//  Read in color-kanji list.
//
  HANDLE handle;
  unsigned long done;
  if (!colorkanji_list) {
    if (INVALID_HANDLE_VALUE != (handle = jfc_config.open(COLORKANJI_NAME,OPEN_READ))) {
      i = GetFileSize(handle,NULL)/sizeof(KANJI);
      if (colorkanji_list = (KANJI *) malloc(i*sizeof(KANJI))) {
        ReadFile (handle,colorkanji_list,i*sizeof(KANJI),&done,NULL);
        colorkanji_size = i;
      }
    }
  }
  CloseHandle (handle);
  return (false);
}



// ### May want to implement a binary search on the color-kanji list.
// ### Direct one-to-one mapping for the color kanji would be better.






