//-------------------------------------------------------------------//
//                                                                   //
//  JFC Copyright (C) Glenn Rosenthal, 1999,2000,2001.               //
//  All rights reserved.                                             //
//                                                                   //
//-------------------------------------------------------------------//

//-------------------------------------------------------------------
//
//  This modlues main function is to deal with the system fonts.  
//  There are three fonts used in the system currently.  These are
//  the kana/kanji font.  The user font used in the main edit window,
//  and the system font used in controls, and in the status bar.  
//  Various useful parameters are caclated buy the font system for use
//  by other routines.  The system currenly allows for changes in the 
//  fonts, buf these are relativly limited.  
//
#ifndef jfc_font_h
#define jfc_font_h

//-------------------------------------------------------------------
//
//  Exported definitons.
//
#ifndef SHIFTJIS_CHARSET            // This constaint is only defined in Win CE 2.11 and 
  #define SHIFTJIS_CHARSET 0x80     //   up.  We support TrueType fonts for all versions, but
#endif  SHIFTJIS_CHARSET            //   must check for Japanese fonts.

typedef unsigned short KANJI;       // Type for a kanji/kana character.

//-------------------------------------------------------------------
//
//  Character type macros (from jwp_inpt.h)
//

#define BASE_KATAKANA   0x2500      // Base value for katakana characters.
#define BASE_HIRAGANA   0x2400      // Base value for hiragana characters.
#define BASE_JASCII     0x2300      // Page for JASCII
#define BASE_KANJI      0x3000      // Value of first actual kanji

#define KANJI_SPACE     0x2121      // Space character in kanji.
#define KANJI_SLASH     0x213F      // Forward slash (Japanese).
#define KANJI_COMMA     0x2122      // Cama (Japanese)
#define KANJI_DOT       0x2126      // Center dot (Japanese)
#define KANJI_ASTRIC    0x2176      // Astric (Japanese)
#define KANJI_ARROW     0x222A      // Horizontal arrow (use in user conversion editor)
#define KANJI_TILDE     0x2141      // ~ (Japanese)
#define KANJI_DASH      0x213D      // - (Japanese, not long vawel, just a dash)
#define KANJI_LONGVOWEL 0x213C      // - (Japanese)
#define KANJI_LPARAN    0x214A      // ( (Japanese)
#define KANJI_RPARAN    0x214B      // ) (Japanese)
#define KANJI_LBRACKET  0x215A      // [ (heavy Japanese)
#define KANJI_RBRACKET  0x215B      // ] (heavy Japanese)   
#define KANJI_BAD       0x2223      // Knaji used for invalid kanji (this is a solid box).
#define KANJI_a         0x2361      // Full width small a
#define KANJI_A         0x2341      // Full width capital A
#define KANJI_k         0x236B      // Full width small k
#define KANJI_K         0x234B      // Full width captical k
#define KANJI_x         0x2378      // Full width small x
#define KANJI_X         0x2358      // Full width capital X
#define KANJI_CROSS     0x215F      // Cross symbol useed in prefix of known kanji lines.
#define KANJI_PLUS      0x215C      // Plus sign (sub-card)
#define KANJI_STAR      0x2176      // Astric 
#define KANJI_HASH      0x2174      // Full width #
#define KANJI_AND       0x2175      // Full width &

#define ISCRLF(ch)     (((ch) == '\n') || ((ch) == '\r'))               // Is this a cr or lf 
#define ISSPACE(ch)    ((ch == ' ') || (ch == '\t') || (ch == 0x2121))  // Is this a space.
#define ISASCII(ch)    (((ch) & 0x7f00) == 0)                           // Is this any ASCII character
#define ISJIS(ch)      (((ch) & 0x7f00) != 0)                           // Is this a kanji or kana
#define ISHIRAGANA(ch) (((ch) & 0x7f00) == BASE_HIRAGANA)               // Is this hiragana
#define ISKATAKANA(ch) (((ch) & 0x7f00) == BASE_KATAKANA)               // Is this katakana
#define ISKANA(ch)     (ISHIRAGANA(ch) || ISKATAKANA(ch))               // Is this a kana
#define ISJASCII(ch)   (((ch) & 0x7f00) == BASE_JASCII)
#define ISKANJI(ch)    (((ch)         ) >= BASE_KANJI)

//-------------------------------------------------------------------
//
//  class KANJI_font.
//
//  This class contains information related direclty to a specific
//  kanji font.
//

//
//  This is a virtual font class.  This class cannot be used directly, but must be used
//  to define a veritual class for the type of font that you want.  Because the font 
//  parameters differ so radically, different open versions for each font type are used.
//  The current routine open_bitmap() and open_font() provide uniformed ways to open the
//  fonts.
//
class KANJI_font {
public:
  short width,height;       // height and witdth of font
  short leading;            // Vertial gaps between lines
  short spacing;            // Horizontal gaps between characters.
  short truetype;           // Indicates this is a TrueType font.
  void  virtual draw       (HDC hdc,int jis,int x,int y) = 0;           // Render character
  void  virtual fill       (HDC hdc,int jis,RECT *rect) = 0;            // Used to make big kanji.
  void          find_color (int jis,HDC hdc,COLORREF &color);           // Find a kanji.  Load the kanji from disk if necessary
  void          remove     (void);                                      // Deallocate this font and clean up.
  int   virtual jis_index  (int jis) = 0;                               // Convert jis code to font code.
private:
  void  virtual close      (void) = 0;                                  // Close the font
};

//  
//  This dirived class processes bitmaped fonts
//
class BITMAP_KANJI_font : public KANJI_font {
public:
  BITMAP_KANJI_font (void);
  void  draw        (HDC hdc,int jis,int x,int y);                      // Render character
  void  fill        (HDC hdc,int jis,RECT *rect);                       // Used to make big kanji.
  int   jis_index   (int jis);                                          // Convert jis code to font code.
  int   open        (TCHAR *name,int cache);                            // Open the font.
private:
  void  close       (void);                                             // Closes a font.
  int   find_kanji  (int index,HDC hdc,COLORREF &color);                // Find a kanji.  Load the kanji from disk if necessary

  short   holes;                // Indicates font has holes.
  short   hshift;               // Shift used in redering characters.
  short   bmsize;               // Size of bitmap in memory bytes (per character).
  short   bmfilesize;           // Size of bitmap in file bytes (per character).
  char   *bitmaps;              // Font's bitmap data
  KANJI  *cache;                // Font cache.
  short   index;                // Next bitmap to replace in the font cache.
  HDC     hdcmem;               // Memory HDC used for rendering the bitmaps.
  HANDLE  file;                 // Pointer to file (used only for cached fonts).
#ifndef WINCE                   // CE does not support SetBitmapBits so cannot use this.
  HBITMAP hbitmap;				// Bitmap for bit operations
#endif
};

//
//  This dirived class processes TrueType fonts.
//
class TRUETYPE_KANJI_font : public KANJI_font {
public:
  TRUETYPE_KANJI_font (void);
  void  draw          (HDC hdc,int jis,int x,int y);                    // Render character
  void  fill          (HDC hdc,int jis,RECT *rect);                     // Used to make big kanji.
  int   jis_index     (int jis);                                        // Convert jis code to font code.
  int   open          (HDC hdc,TCHAR *name,int cache);                  // Open the font.
private:
  void  close         (void);                                           // Closes a font.
  HFONT   font;                 // TrueType font structure.
  short   hshift;               // Shift used in redering characters.
  short   vshift;               // Vertical shift for rendering characters
#ifndef WINCE
  char   *cmap;                 // Copy of the cmap structure for this font.
  char   *gsub;                 // GSUB structre for this font, used for viertical glyph replacemnt
  ushort  count;                // Number of entries in the cmap subtable.
  ushort *start;                // Start table for ranges                                
  ushort *end;                  // End table for ranges
  ushort *delta;                // Delta table for displacements
  ushort *offset;               // Offest table for entriesinto the glyph table
  ushort *glyph;                // Glyph table.
  ushort  vcount;               // Number of vertical replacement glyphs
  ushort *from;                 // Glphys to be replaced.
  ushort *to;                   // replacement glyphs.
#endif WINCE
};

//-------------------------------------------------------------------
//
//  class JFC_font.
//

//
//  This is the main class for font work.  Each JFC_font object contains all
//  the information necessary to render text in a particular font.  All spacings
//  and other settings are contained in the class as well as the ASCII font 
//  metrics.  
//
//  Each JFC_font object really constists of a kanji font and an ASCII font pair,
//  plus all the support information.  
//
//  JFC_font objects can contain dupicates of open fonts.  This makes things more
//  efficient.  They are flaged by the duplicate member.
//
class JFC_font {
public:
  short height;                             // Kanji font height
  short hwidth;                             // Kanji nominal character-character spcaking.
  short rheight;                            // Top of rectange around kanji font (used a lot in redraw, and selecting).
  short vheight;                            // Kanji nominal line-line spacing.
  short lheight;                            // Height used in list boxes.
  short loffset;                            // Offset for rending line of text in list box.  (Only used once, but pre-calculated for speed)
  short cheight;                            // Height of caret (cursor)
  short vspace;                             // Vertical extra space measure (framing around window).
  short hspace;                             // Horizontal extra space meauser (framing around window).
  short x_offset;                           // Horizontal offset for rendering text (leaves a small borer on left side). [Parameter used by the JFC_file class]
  short y_offset;                           // Vertical offset for rendering of text from top of screen. [Parameter used by the JFC_file class]
  byte  duplicate;                          // Indicates that the font is actually a duplicate of another font, so don't deallocate.
  HFONT       ascii;                        // Ascii font.
  KANJI_font *kanji;                        // Kanji font.

  void close      (void);                   // Close the font.
  void copy       (class JFC_font *font);   // Duplicates a particular font.
  int  hadvance   (int x,int ch);           // Advance a cursor horizontaly by a single character.
  int  open       (TCHAR *name,int size,int cache,HDC hdc); // Open and initialize the font.

  inline JFC_font (void) { ascii = NULL; kanji = NULL; duplicate = false; }
private: 
  short widths[256];                        // Holds the char widths in pixels, for the pirmary ASCII.
};

//-------------------------------------------------------------------
//
//  Exported data
//

extern class JFC_font sys_font;             // Instance of the font data.
extern class JFC_font kanji_font;           // Kanji display font.
extern class JFC_font kana_font;            // Kana display font.
extern class JFC_font imi_font;             // Meaning display font.
extern short sysfont_height;                // Height of the system font (used for toolbars).

//-------------------------------------------------------------------
//
//  Exported routines.
//
void        free_fonts       (int full);    // Cleanup routine to close all open fonts.
KANJI_font *get_bigfont      (RECT *rect);  // Routine to get the big font.
int         in_colorkanji    (int ch);      // Determine if a kanji is in the color kanji list.
int         initialize_fonts (void);        // Setup the fonts system and intiailize.

#endif jfc_font_h


